# -*- coding: utf-8 -*-
#    pathpy is an OpenSource python package for the analysis of time series data
#    on networks using higher- and multi order graphical models.
#
#    Copyright (C) 2016-2017 Ingo Scholtes, ETH Zürich
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#    Contact the developer:

#    E-mail: ischoltes@ethz.ch
#    Web:    http://www.ingoscholtes.net
import sys
from collections import defaultdict
import bisect
import datetime
from time import mktime

import numpy as _np

from pathpy.utils import Log, Severity


class TemporalNetwork:
    """
      This class represents a sequence of time-stamped edges.
       Instances of this class can be used to generate path statistics
       based on the time-respecting paths resulting from a given maximum
       time difference between consecutive time-stamped edges.
    """

    def __init__(self, tedges=None):
        """Constructor that generates a temporal network instance.

        Parameters
        ----------
        tedges:
            an optional list of directed time-stamped edges from which to construct a
            temporal network instance. For the default value None an empty temporal
            network will be created.
        """
        # A list of time-stamped edges of this temporal network
        self.tedges = []

        # A list of nodes of this temporal network
        self.nodes = []

        # A dictionary storing all time-stamped links, indexed by time-stamps
        self.time = defaultdict(lambda: list())

        # A dictionary storing all time-stamped links, indexed by time and target node
        self.targets = defaultdict(lambda: dict())

        # A dictionary storing all time-stamped links, indexed by time and source node
        self.sources = defaultdict(lambda: dict())

        # A dictionary storing time stamps at which links (v,*;t) originate from node v
        self.activities = defaultdict(lambda: list())

        # A dictionary storing sets of time stamps at which links (v,*;t) originate from
        # node v
        # Note that the insertion into a set is much faster than repeatedly checking
        # whether an element already exists in a list!
        self.activities_sets = defaultdict(lambda: set())

        # An ordered list of time-stamps
        self.ordered_times = []

        nodes_seen = defaultdict(lambda: False)

        if tedges is not None:
            Log.add('Building index data structures ...')

            for e in tedges:
                self.activities_sets[e[0]].add(e[2])
                self.time[e[2]].append(e)
                self.targets[e[2]].setdefault(e[1], []).append(e)
                self.sources[e[2]].setdefault(e[0], []).append(e)
                if not nodes_seen[e[0]]:
                    nodes_seen[e[0]] = True
                if not nodes_seen[e[1]]:
                    nodes_seen[e[1]] = True
            self.tedges = tedges
            self.nodes = list(nodes_seen.keys())

            Log.add('Sorting time stamps ...')

            self.ordered_times = sorted(list(self.time.keys()))
            for v in self.nodes:
                self.activities[v] = sorted(self.activities_sets[v])
            Log.add('finished.')

    @classmethod
    def from_sqlite(cls, cursor, directed=True, timestamp_format='%Y-%m-%d %H:%M:%S'):
        """Reads time-stamped links from an SQLite cursor and returns a new instance of
        the class TemporalNetwork. The cursor is assumed to refer to a table that
        minimally has three columns

                source target time

        and where each row refers to a directed link. Time stamps can be integers,
        or strings to be converted to UNIX time stamps via a custom timestamp format.
        For this, the python function datetime.strptime will be used.

        Important: Since columns are accessed by name this function requires that a
        row factory object is set for the SQLite connection prior to cursor creation,
        i.e. you should set

                connection.row_factory = sqlite3.Row

        Parameters
        ----------
        cursor:
            The SQLite cursor to fetch rows
        directed: bool
        timestamp_format: str
            used to convert string timestamps to UNIX timestamps. This parameter is
            ignored, if the timestamps are digit types (like a simple int).

        Returns
        -------

        """
        tedges = []

        assert cursor.connection.row_factory, \
            'Cannot access columns by name. Please set ' \
            'connection.row_factory = sqlite3.Row before creating DB cursor.'

        if not directed:
            Log.add('Retrieving undirected time-stamped links ...')
        else:
            Log.add('Retrieving directed time-stamped links ...')

        for row in cursor:
            # r = sqlite3.Row(row)
            timestamp = row['time']
            assert isinstance(timestamp, int) or isinstance(timestamp, str), \
                'Error: pathpy only supports integer or string timestamps'
            # if the timestamp is a number, we use this
            if isinstance(timestamp, int):
                t = timestamp
            else:
                # if it is a string, we use the timestamp format to convert it to
                # a UNIX timestamp
                x = datetime.datetime.strptime(timestamp, timestamp_format)
                t = int(mktime(x.timetuple()))
            tedges.append((str(row['source']), str(row['target']), t))
            if not directed:
                tedges.append((str(row['target']), str(row['source']), t))

        return cls(tedges=tedges)

    @classmethod
    def read_file(cls, filename, sep=',', directed=True,
                  timestamp_format='%Y-%m-%d %H:%M:%S', maxlines=sys.maxsize):
        """
        Reads time-stamped links from a file and returns a new instance of the class
        TemporalNetwork. The file is assumed to have a header

                source target time

        where columns can be in arbitrary order and separated by arbitrary characters.
        Each time-stamped link must occur in a separate line and links are assumed to be
        directed.

        The time column can be omitted and in this case all links are assumed to occur
        in consecutive time stamps (that have a distance of one). Time stamps can be
        simple integers, or strings to be converted to UNIX time stamps via a custom
        timestamp format. For this, the python function datetime.strptime will be used.

        Parameters
        ----------
        filename: str
            path of the file to read from
        sep: str
            the character that separates columns (default ',')
        directed: bool
            whether to read edges as directed (default True)
        timestamp_format: str
            used to convert string timestamps to UNIX timestamps. This parameter is
            ignored, if timestamps are digit types (like a simple int).
            The default is '%Y-%m-%d %H:%M'
        maxlines: int
            limit reading of file to a given number of lines (default sys.maxsize)

        Returns
        -------

        """
        assert (filename != ''), 'Empty filename given'

        # Read header
        with open(filename, 'r') as f:
            tedges = []

            header = f.readline()
            header = header.split(sep)

            # If header columns are included, arbitrary column orders are supported
            time_ix = -1
            source_ix = -1
            target_ix = -1
            for i in range(len(header)):
                header[i] = header[i].strip()
                if header[i] == 'node1' or header[i] == 'source':
                    source_ix = i
                elif header[i] == 'node2' or header[i] == 'target':
                    target_ix = i
                elif header[i] == 'time' or header[i] == 'timestamp':
                    time_ix = i            

            assert (source_ix >= 0 and target_ix >= 0), \
                "Detected invalid header columns: %s" % header

            if time_ix < 0:  # pragma: no cover
                Log.add('No time stamps found in data, assuming consecutive links',
                        Severity.WARNING)

            if not directed:
                Log.add('Reading undirected time-stamped links ...')
            else:
                Log.add('Reading directed time-stamped links ...')

            line = f.readline()
            n = 1
            while line and n <= maxlines:
                fields = line.rstrip().split(sep)
                try:
                    if time_ix >= 0:
                        timestamp = fields[time_ix]
                        # if the timestamp is a number, we use this
                        if timestamp.isdigit():
                            t = int(timestamp)
                        else:
                            # if it is a string, we use the timestamp format to convert
                            # it to a UNIX timestamp
                            x = datetime.datetime.strptime(timestamp, timestamp_format)
                            t = int(mktime(x.timetuple()))
                    else:
                        t = n
                    if t >= 0 and fields[source_ix] != '' and fields[target_ix] != '':
                        tedge = (fields[source_ix], fields[target_ix], t)
                        tedges.append(tedge)
                        if not directed:
                            tedges.append((fields[target_ix], fields[source_ix], t))
                    else:  # pragma: no cover
                        s_line = line.strip()
                        if fields[source_ix] == '' or fields[target_ix] == '':
                            msg = 'Empty node in line {0}: {1}'.format(n+1, s_line)
                        else: 
                            msg = 'Negative timestamp in line {0}: {1}'.format(n+1, s_line)
                        Log.add(msg, Severity.WARNING)
                except (IndexError, ValueError):  # pragma: no cover
                    s_line = line.strip()
                    msg = 'Malformed line {0}: {1}'.format(n+1, s_line)
                    Log.add(msg, Severity.WARNING)
                line = f.readline()
                n += 1
        # end of with open()

        return cls(tedges=tedges)

    def write_file(self, filename, sep=','):
        """Writes the time-stamped edge list of this temporal network instance as CSV file

        Parameters
        ----------
        filename: str
            name of CSV file to save data to
        sep: str
            character used to separate columns in generated CSV file

        Returns
        -------

        """
        msg = 'Writing {0} time-stamped edges to file {1}'.format(self.ecount(), filename)
        Log.add(msg, Severity.INFO)
        with open(filename, 'w+') as f:
            f.write('source' + sep + 'target' + sep + 'time' + '\n')
            for time in self.ordered_times:
                for (v, w, t) in self.time[time]:
                    f.write(str(v) + sep + str(w) + sep + str(t)+'\n')


    def filter_nodes(self, nodes):
        """Returns a copy of the temporal network where time-stamped edges are filtered 
        according to a given set of nodes.

        Parameters
        ----------
        node_filter: iterable
            a list or set of nodes that shall be included in the returned temporal network

        Returns
        -------
        """
        def edge_filter(v, w, t):
            if v in nodes and w in nodes: 
                return True
            return False

        return self.filter_edges(edge_filter)

    def filter_edges(self, edge_filter):
        """Returns a copy of the temporal network where time-stamped edges are filtered 
        according to a given filter expression. This can be used, e.g., to create time 
        slice networks by filtering edges within certain time windows, or to reduce a 
        temporal network to interactions between a subset of nodes.

        Parameters
        ----------
        edge_filter: callable
            an arbitrary filter function of the form filter_func(v, w, time) that returns
            True for time-stamped edges that shall pass the filter, and False for time-stamped
            edges that shall be filtered out.

        Returns
        -------

        """
        Log.add('Starting filtering ...', Severity.INFO)
        new_t_edges = []

        for (v, w, t) in self.tedges:
            if edge_filter(v, w, t):
                new_t_edges.append((v, w, t))

        n_filtered = self.ecount() - len(new_t_edges)
        msg = 'finished. Filtered out {} time-stamped edges.'.format(n_filtered)
        Log.add(msg,  Severity.INFO)

        return TemporalNetwork(tedges=new_t_edges)

    def add_edge(self, source, target, ts, directed=True):
        """Adds a time-stamped edge (source,target;time) to the temporal network.
        Unless specified otherwise, time-stamped edges are assumed to be directed.

        Parameters
        ----------
        source:
            name of the source node of a directed, time-stamped link
        target:
            name of the target node of a directed, time-stamped link
        ts: int
            time-stamp of the time-stamped link
        directed: bool

        Returns
        -------

        """
        e = (source, target, ts)
        self.tedges.append(e)
        if source not in self.nodes:
            self.nodes.append(source)
        if target not in self.nodes:
            self.nodes.append(target)

        # Add edge to index structures
        self.time[ts].append(e)
        self.targets[ts].setdefault(target, []).append(e)
        self.sources[ts].setdefault(source, []).append(e)

        if ts not in self.activities[source]:
            self.activities[source].append(ts)
            self.activities[source].sort()

        # Maintain order of time stamps
        index = bisect.bisect_left(self.ordered_times, ts)
        # add if ts is not already in list
        if index == len(self.ordered_times) or self.ordered_times[index] != ts:
            self.ordered_times.insert(index, ts)

        # make edge undirected by adding another directed edge
        if not directed:
            self.add_edge(target, source, ts)

    def vcount(self):
        """Returns the number of vertices in the temporal network.
        This number corresponds to the number of nodes in the (first-order)
        time-aggregated network.
        """

        return len(self.nodes)

    def ecount(self):
        """Returns the number of time-stamped edges (u,v;t) in the temporal network.
        This number corresponds to the sum of link weights in the (first-order)
        time-aggregated network.
        """

        return len(self.tedges)

    def observation_length(self):
        """Returns the length of the observation time in time units."""

        return max(self.ordered_times)-min(self.ordered_times)

    def inter_event_times(self):
        """
        Returns an array containing all time differences between any
        two consecutive time-stamped links (involving any node)
        """
        time_diffs = []
        for i in range(1, len(self.ordered_times)):
            time_diffs += [self.ordered_times[i] - self.ordered_times[i-1]]
        return _np.array(time_diffs)

    def inter_path_times(self):
        """Returns a dictionary which, for each node v, contains all time differences
        between any time-stamped link (*,v;t) and the next link (v,*;t') (t'>t) in the
        temporal network
        """
        ip_times = defaultdict(list)
        for e in self.tedges:
            # Get target v of current edge e=(u,v,t)
            v = e[1]
            t = e[2]

            # Get time stamp of link (v,*,t_next)
            # with smallest t_next such that t_next > t
            i = bisect.bisect_right(self.activities[v], t)
            if i != len(self.activities[v]):
                ip_times[v].append(self.activities[v][i]-t)
        return ip_times

    def summary(self):
        """
        Returns a string containing basic summary statistics of this temporal network
        """

        summary = ''

        summary += 'Nodes:\t\t\t' + str(self.vcount()) + '\n'
        summary += 'Time-stamped links:\t' + str(self.ecount()) + '\n'
        if self.vcount() > 0:
            summary += 'Links/Nodes:\t\t' + str(self.ecount()/self.vcount()) + '\n'
        else:
            summary += 'Links/Nodes:\t\tN/A\n'
        if len(self.ordered_times) > 1:
            min_o = min(self.ordered_times)
            max_o = max(self.ordered_times)
            obs_len = max_o - min_o
            n_stamps = len(self.ordered_times)
            summary += 'Observation period:\t[{}, {}]\n'.format(min_o, max_o)
            summary += 'Observation length:\t {} \n'.format(obs_len)
            summary += 'Time stamps:\t\t {} \n'.format(n_stamps)

            d = self.inter_event_times()
            mean_d = _np.mean(d)

            summary += 'Avg. inter-event dt:\t {}\n'.format(mean_d)
            summary += 'Min/Max inter-event dt:\t {}/{}'.format(min(d), max(d))

        return summary

    def __str__(self):
        """Returns the default string representation of this temporal network instance.
        """
        return self.summary()

    def shuffle_edges(self, l=0, with_replacement=False, window_splits=None,
                      maintain_undirected=True):
        """Generates a shuffled version of the temporal network in which edge statistics
        (i.e. the frequencies of time-stamped edges) and inter-event time  statistics are
        preserved, while all order correlations are destroyed by randomly reshuffling the
        time stamps of links.

        Parameters
        ----------
        l: int
            For the default value l=0, the length of the original temporal network is
            used.
        with_replacement: bool
            Whether or not the sampling of new time-stamped edges should be with
            replacement (default False). If False, the exact edge frequencies and
            inter-event time statistics in the original network will be preserved.
        window_splits: list
            a list of time stamps that separate shuffling windows. E.g. specifying
            window_splits = [7,14,21] will separately shuffle edges within intervals
            [min_timestamp,7], (7,14], (14,21], (21,max_timestamp] (default None). The
            number of edges l to generate applies separately for each time window. For
            l=0, the original number of edges in each time window will be used.
        maintain_undirected: bool
            if True, two directed edges (a,b,t) (b,a,t) occurring at the same time will be
            treated as a single undirected edge, i.e. both are reassigned to a different
            time stamp at once (default True). This ensures that undirected edges are
            preserved as atomic objects.

        Returns
        -------

        """
        tedges = []

        if window_splits is None:
            window_splits = [max(self.time)]
        else:
            window_splits.append(max(self.time))

        window_min = min(self.time)-1
        for window_max in window_splits:

            timestamps = []
            edges = []
            for e in self.tedges:
                if window_min < e[2] <= window_max:
                    timestamps.append(e[2])
                    edges.append(e)

            if l == 0:
                l = len(edges)
            if with_replacement:  # sample l edges with replacement
                for i in range(l):
                    # Pick random link
                    edge = edges[_np.random.randint(0, len(edges))]
                    # Pick random time stamp
                    time = timestamps[_np.random.randint(0, len(timestamps))]
                    # Generate new time-stamped link
                    tedges.append((edge[0], edge[1], time))
            else:
                # shuffle edges while avoiding multiple identical edges at same time stamp
                i = 0
                while i < l:
                    # Pick random link
                    edge = edges.pop(_np.random.randint(0, len(edges)))

                    # Pick random time stamp
                    time = timestamps.pop(_np.random.randint(0, len(timestamps)))
                    rewired = False

                    # for undirected edges, rewire both directed edges at once
                    if maintain_undirected and (edge[1], edge[0], edge[2]) in edges:

                        # check whether one of the time-stamped edges already exists
                        if (
                                (edge[0], edge[1], time) not in tedges and
                                (edge[1], edge[0], time) not in tedges
                        ):
                            tedges.append((edge[0], edge[1], time))
                            tedges.append((edge[1], edge[0], time))
                            edges.remove((edge[1], edge[0], edge[2]))
                            rewired = True

                    # rewire a single directed edge individually
                    elif (edge[0], edge[1], time) not in tedges:
                        tedges.append((edge[0], edge[1], time))
                        rewired = True

                    # edge could not be rewired to the chosen time stamp, so re-append
                    # both to the list for future sampling
                    if not rewired:
                        edges.append(edge)
                        timestamps.append(time)
                    else:
                        i += 1

            window_min = window_max

        # Generate temporal network
        t = TemporalNetwork(tedges=tedges)

        # Fix node order to correspond to original network
        t.nodes = self.nodes
        return t

    def write_tikz(self, filename, dag=True, angle=20, layer_dist='0.3cm',
                   split_directions=True):
        """Generates a tex file that can be compiled to a time-unfolded representation of
         the temporal network.

        Parameters
        ----------
        filename: str
            the name of the tex file to be generated.
        dag: bool
            whether or not to draw the unfolded network as a directed acyclic graph,
            in which a link (v,w,t) connects node-time elements (v_t, w_t+1) (default
            True). If False, a simple sequence of links will be generated.
        angle: float
            the angle of curved edges
        layer_dist: str
            LaTex distance parameter string specifying the distance between adjacent
            node-time elements (default '0.3cm')
        split_directions: bool
            whether or not the curve angle of edges shall be split depending on
            direction (default True) If this is set to True, arrows from left to right
            bend upwards, while arrows from right to left bend downwards This helps
            readability in temporal networks with multiple edges per time step. For
            temporal networks with single edges per time, False is recommended.

        Returns
        -------

        """
        import os as _os

        output = [
            '\\documentclass{article}',
            '\\usepackage{tikz}',
            '\\usepackage{verbatim}',
            '\\usepackage[active,tightpage]{preview}',
            '\\PreviewEnvironment{tikzpicture}',
            '\\setlength\\PreviewBorder{5pt}%',
            '\\usetikzlibrary{arrows}',
            '\\usetikzlibrary{positioning}',
            '\\renewcommand{\\familydefault}{\\sfdefault}',
            '\\begin{document}',
            '\\begin{center}',
            '\\newcounter{a}',
            "\\begin{tikzpicture}[->,>=stealth',auto,"
            "scale=1, every node/.style={scale=1}]",
            "\\tikzstyle{node} = [fill=lightgray,text=black,circle]",
            "\\tikzstyle{v} = [fill=lightgray,draw=black,"
            "text=white,circle,minimum size=0.5cm]",
            "\\tikzstyle{dst} = [fill=lightgray,text=black,circle]",
            "\\tikzstyle{lbl} = [text=black,circle]"]

        last = ''

        for n in _np.sort(self.nodes):
            if last == '':
                node = r"\node[lbl]  ({n}-0) {{\bf \Huge {n} }};".format(n=n)
            else:
                node_fmt = r"\node[lbl, right=0.4cm of {last}-0] ({n}-0) " \
                           r"{{\bf \Huge {n} }};"
                node = node_fmt.format(last=last, n=n)

            output.append(node)
            last = n

        output.append("\\setcounter{a}{0}\n")
        min_t = min(self.ordered_times)
        max_t = max(self.ordered_times)
        if dag:
            dag_num = r'\foreach \number in {{ {min_t}, ..., {max_t} }} {{'
            output.append(dag_num.format(min_t=min_t, max_t=max_t+1))
        else:
            dag_num = r'\foreach \number in {{ {min_t}, ..., {max_t} }} {{'
            output.append(dag_num.format(min_t=min_t, max_t=max_t))
        output.append("\\setcounter{a}{\\number}")
        output.append("\\addtocounter{a}{-1}")
        output.append("\\pgfmathparse{\\thea}")

        for n in _np.sort(self.nodes):
            layer_fm = r"\node[v,below={layer} of {n}-\pgfmathresult] ({n}-\number) {{}};"
            output.append(layer_fm.format(layer=layer_dist, n=n))

        first_node = _np.sort(self.nodes)[0]
        fmt = r'\node[lbl,left=0.4cm of {f_node}-\number] (col-\pgfmathresult) ' \
              r'{{ \selectfont {{ \bf \Huge \number}} }};'

        output.append(fmt.format(f_node=first_node))
        output.append(r"}")
        output.append(r"\path[->,line width=2pt]")
        # draw only directed edges
        for ts in self.ordered_times:
            for edge in self.time[ts]:
                if dag:
                    edge_str = r'({}-{}) edge ({}-{})'.format(edge[0], ts, edge[1], ts+1)
                    output.append(edge_str)
                else:
                    if (edge[1], edge[0], ts) not in self.time[ts]:
                        bend_direction = 'right'
                        if not split_directions and edge[0] < edge[1]:
                            bend_direction = 'left'
                        edge_fmt = r'({}-{}) edge[bend {} = {}] ({}-{})'
                        edge_str = edge_fmt.format(edge[0], ts, bend_direction, angle,
                                                   edge[1], ts)
                        output.append(edge_str)
        output.append(";\n")

        # separately draw undirected edges if we don't output a DAG
        if not dag:
            output.append("\\path[-,line width=2pt]\n")
            for ts in self.ordered_times:
                for edge in self.time[ts]:
                    if (edge[1], edge[0], ts) in self.time[ts]:
                        # admittedly, this is an ugly trick: I avoid keeping state on
                        # which of the directed edges has been drawn already as an
                        # undirected edge, by simply drawing them twice in the same way
                        #  :-)
                        s = max(edge[0], edge[1])
                        t = min(edge[0], edge[1])
                        edge = "({s}-{ts}) edge[bend right={angle}] ({t}={ts})\n"
                        output.append(edge.format(s=s, ts=ts, angle=angle, t=t))
            output.append(";\n")
        output.append("\\end{tikzpicture} \n"
                      "\\end{center} \n"
                      "\\end{document}")

        # create directory if necessary to avoid IO errors
        directory = _os.path.dirname(filename)
        if directory != '':
            if not _os.path.exists(directory):
                _os.makedirs(directory)

        with open(filename, "w") as tex_file:
            tex_file.writelines('\n'.join(output))


    def _to_html(self, width=600, height=600, ms_per_frame=50, ts_per_frame=20, radius=6,
                look_behind=1500, look_ahead=150, template_file=None, **kwargs):
        """
        Generates an html snippet with an interactive d3js visualisation of the
        temporal network. This function can be used to embed interactive visualisations
        into a jupyter notebook, or to export stand-alone html visualisations based on
        a customizable template.

        Parameters
        ----------
        width:  int
            the width of the div that contains the visualisation (default 600)
        height: int
            the height of the div that contains the visualisation (default 600)
        ms_per_frame: int
            how many milliseconds each frame of the visualisation shall be displayed.
            The inverse of this value gives the framerate of the resulting visualisation.
            The default value of 100 yields a framerate of 10 fps
        ts_per_frame: int
            how many timestamps in the temporal network shall be displayed in every frame
            of the visualisation (default 1). For the default value of 1, each timestamp
            is shown in a new frame. For higher values, multiple timestamps will be aggregated
            in a single frame.
        radius: int
            radius of nodes in the visualisation. Unfortunately this can only be set via
            the style file starting from SVG2.
        template_file: str
            path to the template file that shall be used to output the html visualisation (default None)
            If a custom-tailored template_file is specified, python's string.Template mechanism is used
            to replace the following strings by JavaScript variables:
                $network_data:  replaced by a JSON dictionary that contains nodes and time-stamped links
                $width: width of the DIV to be generated
                $height: height of the DIV to be generated
                $msperframe: see above
                $tsperframe: see above
                $div_id: a unique ID for the generated DIV. This is important when including
                    multiple visualisations into a single output file (e.g. in multiple jupyter cells)
            If this is set to None (default value), a default template provided by pathpy will be used.
        use_requirejs: bool
            whether or not the generated html shall import d3js via the requirejs framework
            (default True). The use of requirejs is needed to include html inside a jupyter
            notebook. For the generation of stand-alone files, the value should be set to
            False. This parameter will be ignored when using a custom templatefile
        **kwargs: keyword args
            arbitrary key-value pairs, that will be exported (via json.dumps) to the corresponding
            placeholder values in a custom template-file. As an example, if the template file contains
            JavaScript code like x=f($x); and console.log($y);, and we set parameters
            write_html(..., x=42, y='knockknock'), the exported JavaScript will contain x=f(42) and
            console.log('knockknock').
        Returns
        -------

        """
        import json
        import os
        from string import Template

        import string
        import random

        # auto-adjust simulation speed to temporal characteristics
        if ts_per_frame == 0:
            d = self.inter_event_times()
            avg_ts_bw_interactions = _np.mean(d)
            fps = 1000.0/float(ms_per_frame)
            x = avg_ts_bw_interactions/fps
            # set time scale so that we expect 5 interactions per frame
            ts_per_frame = _np.max([1, int(20 * x)])

        div_id = "".join(random.choice(string.ascii_letters) for x in range(8))

        # prefix nodes starting with number
        def fix_node_name(v):
            new_v = v
            if v[0].isdigit():
                new_v = "n_" + v
            if new_v[0] == '_':
                new_v = "n_" + v
            if '-' in new_v:
                new_v = new_v.replace('-', '_')
            return new_v

        network_data = {
            'nodes': [{'id': fix_node_name(v), 'group': 1} for v in self.nodes],
            'links': [{'source': fix_node_name(s),
                       'target': fix_node_name(v),
                       'value': 1,
                       'time': t} for s, v, t in self.tedges
                      ]
        }

        # use standard template if no custom template is specified
        if template_file is None:
            module_dir = os.path.dirname(os.path.realpath(__file__))
            html_dir = os.path.join(module_dir, os.path.pardir, 'html_templates')
            template_file = os.path.join(html_dir, 'tempnet.html')            

        with open(template_file) as f:
            html_str = f.read()

        default_args = {
            'network_data': json.dumps(network_data),
            'width': width,
            'height': height,
            'div_id': div_id,
            'msperframe': ms_per_frame,
            'tsperframe': ts_per_frame,
            'radius': radius,
            'look_ahead': look_ahead,
            'look_behind': look_behind
        }

        # replace all placeholders in the template
        html = Template(html_str).substitute({**default_args, **kwargs})

        return html


    def _repr_html_(self, width=800, height=800, msperframe=50, tsperframe=20, radius=6,
            look_behind=1500, look_ahead=150, template_file=None, **kwargs):
        from IPython.core.display import display, HTML
        display(HTML(self._to_html(width, height, msperframe, tsperframe=tsperframe, radius=radius,
            template_file=template_file, **kwargs)))


    def write_html(self, filename, width=800, height=800, msperframe=50, tsperframe=20, radius=6,
            look_behind=1500, look_ahead=150, template_file=None, **kwargs):

        html = self._to_html(width, height, msperframe, tsperframe=tsperframe, radius=radius,
            template_file=template_file, **kwargs)

        # for the inner HTML generated from the default templates, we add the surrounding DOCTYPE and body
        # needed for a stand-alone HTML file
        if template_file is None:
            html = '<!DOCTYPE html>\n<html><body>\n' + html + '</body>\n</html>'
        with open(filename, 'w+') as f:
            f.write(html)



# -*- coding: utf-8 -*-
#    pathpy is an OpenSource python package for the analysis of time series data
#    on networks using higher- and multi order graphical models.
#
#    Copyright (C) 2016-2017 Ingo Scholtes, ETH Zürich
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#    Contact the developer:

#    E-mail: ischoltes@ethz.ch
#    Web:    http://www.ingoscholtes.net
import sys
import itertools as it
import functools as ft
from collections import defaultdict

from pathpy.classes.paths import Paths
from pathpy.utils import Log, Severity
from pathpy import DAG


def remove_repetitions(path):
    """Remove repeated nodes in the path

    Parameters
    ----------
    path

    Returns
    -------

    Examples
    -------
    >>> remove_repetitions((1, 2, 2, 3, 4, 1))
    (1, 2, 3, 4, 1)
    >>> remove_repetitions((1, 2, 2, 2, 3)) == remove_repetitions((1, 2, 2, 3, 3))
    True
    """
    return tuple(p[0] for p in it.groupby(path))


def paths_from_dag(dag, node_mapping=None, max_subpath_length=None, separator=',', repetitions=True):
    """    Extracts pathway statistics from a directed acyclic graph.
    For this, all paths between all roots (zero incoming links)
    and all leafs (zero outgoing link) will be constructed.

    Parameters
    ----------
    dag: DAG
    node_mapping: dict
        can be a simple mapping (1-to-1) or a 1-to-many (a dict with sets as values)
    max_subpath_length: int
    separator: str
        separator to use to separate nodes on paths

    Returns
    -------
    Paths

    """
    # Try to topologically sort the graph if not already sorted
    if node_mapping:
        test_key = list(node_mapping.keys())[0]
        ONE_TO_MANY = isinstance(node_mapping[test_key], set)
    else:
        ONE_TO_MANY = False

    if dag.is_acyclic is None:
        dag.topsort()
    # issue error if graph contains cycles
    if dag.is_acyclic is False:
        Log.add('Cannot extract statistics from a cyclic graph', Severity.ERROR)
        raise ValueError('Cannot extract path statistics from a cyclic graph')
    else:
        # path object which will hold the detected (projected) paths
        p = Paths(separator=separator)
        if max_subpath_length:
            p.max_subpath_length = max_subpath_length
        else:
            p.max_subpath_length = sys.maxsize

        Log.add('Creating paths from directed acyclic graph', Severity.INFO)

        # construct all paths originating from root nodes for 1 to 1
        if not ONE_TO_MANY:
            for s in dag.roots:
                extracted_paths = dag.routes_from_node(s, node_mapping)
                for path in extracted_paths:   # add detected paths to paths object
                    if repetitions:
                        p.add_path_tuple(path, expand_subpaths=False, frequency=(0, 1))
                    else:
                        p.add_path_tuple(remove_repetitions(path), expand_subpaths=False, frequency=(0, 1))
        else:
            path_counter = defaultdict(lambda: 0)
            for root in dag.roots:
                for set_path in dag.routes_from_node(root, node_mapping):
                    for blown_up_path in expand_set_paths(set_path):
                        path_counter[blown_up_path] += 1

            for path, count in path_counter.items():
                if repetitions:
                    p.add_path_tuple(path, expand_subpaths=False, frequency=(0, count))
                else:
                    p.add_path_tuple(remove_repetitions(path), expand_subpaths=False, frequency=(0, count))

        Log.add('Expanding Subpaths', Severity.INFO)
        p.expand_subpaths()
        Log.add('finished.', Severity.INFO)
        return p


def expand_set_paths(set_path):
    """returns all possible paths which are consistent with the sequence of sets

    Parameters
    ----------
    set_path: list
        a list of sets or other iterable

    Examples
    -------
    >>> node_path = [{1, 2}, {2, 5}, {1, 2}]
    >>> list(expand_set_paths(node_path))
    [(1, 2, 1), (2, 2, 1), (1, 5, 1), (2, 5, 1), (1, 2, 2), (2, 2, 2), (1, 5, 2), (2, 5, 2)]
    >>> node_path = [{1, 2}, {5}, {2, 5}]
    >>> list(expand_set_paths(node_path))
    [(1, 5, 2), (2, 5, 2), (1, 5, 5), (2, 5, 5)]


    Yields
    ------
    tuple
        a possible path
    """
    # how many possible combinations are there
    node_sizes = [len(n) for n in set_path]
    num_possibilities = ft.reduce(lambda x, y: x * y, node_sizes, 1)

    # create a list of lists such that each iterator is repeated the number of times
    # his predecessors have completed their cycle
    all_periodics = []
    current_length = 1
    for node_set in set_path:
        periodic_num = []
        for num in node_set:
            periodic_num.extend([num] * current_length)
        current_length *= len(node_set)
        all_periodics.append(periodic_num)

    iterator = [it.cycle(periodic) for periodic in all_periodics]
    for i, elements in enumerate(zip(*iterator)):
        if i >= num_possibilities:
            break
        yield elements



# -*- coding: utf-8 -*-
#    pathpy is an OpenSource python package for the analysis of time series data
#    on networks using higher- and multi order graphical models.
#
#    Copyright (C) 2016-2017 Ingo Scholtes, ETH Zürich
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#    Contact the developer:

#    E-mail: ischoltes@ethz.ch
#    Web:    http://www.ingoscholtes.net

from collections import defaultdict
import numpy as _np
import scipy.sparse as _sparse

from pathpy.utils.exceptions import PathsTooShort
from pathpy.classes.network import Network


class HigherOrderNetwork(Network):
    """
    Instances of this class capture a k-th-order representation
    of path statistics. Path statistics can originate from pathway
    data, temporal networks, or from processes observed on top
    of a network topology.
    """

    def __init__(self, paths, k=1, null_model=False, separator=None):
        """Generates a k-th-order representation based on the given path statistics.

        Parameters
        ----------
        paths: Path
            An instance of class Paths, which contains the path statistics to be used in
            the generation of the k-th order representation
        k: int
            The order of the network representation to generate. For the default case of
            k=1, the resulting representation corresponds to the usual (first-order)
            aggregate network, i.e. links connect nodes and link weights are given by the
            frequency of each interaction. For k>1, a k-th order node corresponds to a
            sequence of k nodes. The weight of a k-th order link captures the frequency
            of a path of length k.
        null_model: bool
            For the default value False, link weights capture the frequencies of paths of length k 
            in the underlying paths object. If True, link weights capture expected frequencies
            under the assumption of independent links (i.e. corresponding to a first-order
            Markov model).
        separator: str
            The separator character to be used in higher-order node names. If this parameter 
            is not specified, the separator character of the underlying paths object will be 
            used.
        """
        assert not null_model or (null_model and k > 1)

        if not (paths.paths.keys() and max(paths.paths.keys()) >= k):
            msg = ('Constructing a model of order %d requires paths of at least length %d, '
                   'found paths of max length %d ' % (k, k, max(paths.paths.keys())))
            raise PathsTooShort(msg)

        super().__init__(directed=True)

        # The order of this HigherOrderNetwork
        self.order = k

        # The paths object used to generate this instance
        self.paths = paths

        # The separator character used to label higher-order nodes.
        # For separator '-', the name of a second-order node will be 'a-b'.
        if separator is None:
            self.separator = paths.separator
        else:
            self.separator = separator

        # NOTE: In a higher-order network, edge weights as well as in- and out
        # weights of nodes are numpy arrays consisting of two weight components [w0, w1].
        # w0 counts the weight of an edge based on its occurrence in a subpaths
        # while w1 counts the weight of an edge based on its occurrence in
        # a longest path. As an illustrating example, consider the single
        # path a -> b -> c. In the first-order network, the weights of edges
        # (a,b) and (b,c) are both (1,0). In the second-order network, the
        # weight of edge (a-b, b-c) is (0,1).
        # Here, we will store these weights (as well as in- and out-degrees in
        # node and edge attributes)

        if k > 1:
            # For k>1 we need the first-order network to generate the null model
            # and calculate the degrees of freedom

            # TODO: For a multi-order model, the first-order network is generated multiple
            # times! Make this more efficient
            g1 = HigherOrderNetwork(paths, k=1)
            g1_node_mapping = g1.node_to_name_map()
            A = g1.adjacency_matrix(include_subpaths=True, weighted=False,
                                    transposed=True)            

        if not null_model:
            # Calculate the frequency of all paths of
            # length k, generate k-order nodes and set
            # edge weights accordingly
            iterator = paths.paths[k].items()

            if k == 0:
                # For a 0-order model, we generate a "dummy" start node
                self.add_node('start', inweight=_np.array([0.0, 0.0]), outweight=_np.array([0.0, 0.0]))
                for key, val in iterator:
                    w = key[0]
                    # add weight val to edge ('start', w)
                    self.add_node(w, inweight=_np.array([0.0, 0.0]), outweight=_np.array([0.0, 0.0]))
                    self.add_edge('start', w, weight=val)
            else:
                for key, val in iterator:
                    # Generate names of k-order nodes v and w
                    v = self.separator.join(key[0:-1])
                    w = self.separator.join(key[1:])
                    self.add_node(v, inweight=_np.array([0.0, 0.0]), outweight=_np.array([0.0, 0.0]))
                    self.add_node(w, inweight=_np.array([0.0, 0.0]), outweight=_np.array([0.0, 0.0]))
                    # add weight val to directed edge (v,w)
                    self.add_edge(v, w, weight=val)
            
                # create all possible higher-order nodes
                if k > 1:

                    nodes = HigherOrderNetwork.generate_possible_paths(g1, k-1)
                    for p in nodes:
                        v = p[0]                
                        for l in range(1, k):
                            v = v + self.separator + p[l]
                        
                        # create nodes and make sure that in- and out-weights are numpy arrays
                        if v not in self.nodes:
                            self.add_node(v, inweight=_np.array([0.0, 0.0]), outweight=_np.array([0.0, 0.0]))

            # Note: For all sequences of length k which (i) have never been observed, but
            #       (ii) do actually represent paths of length k in the first-order
            #       network, we may want to include some 'escape' mechanism along the
            #       lines of (Cleary and Witten 1994)

        else:
            # generate the *expected* frequencies of all possible
            # paths of length k based on independently occurring (first-order) links
            possible_paths = HigherOrderNetwork.generate_possible_paths(g1, k)

            # validate that the number of unique paths corresponds to the sum
            # of entries in A**k
            A_sum = _np.sum(A ** k)
            assert A_sum == len(possible_paths), \
                'Expected {ak} paths but got {re}'.format(ak=A_sum, re=len(possible_paths))

            T = g1.transition_matrix(include_subpaths=True)

            # create nodes and links in k-th-order null model
            for p in possible_paths:
                # create higher-order nodes (a,b,c,...) and (b,c,d,...)
                v = p[0]
                for l in range(1, k):
                    v = v + self.separator + p[l]

                w = p[1]
                for l in range(2, k + 1):
                    w = w + self.separator + p[l]                    

                # create nodes and make sure that in- and out-weights are numpy arrays
                self.add_node(v, inweight=_np.array([0.0, 0.0]), outweight=_np.array([0.0, 0.0]))
                self.add_node(w, inweight=_np.array([0.0, 0.0]), outweight=_np.array([0.0, 0.0]))

                # In the null model, we encode a first-order Markov process in a k-th-order
                # model. For the transition probabilities e.g. (a,b) -> (b,c) in a second-order
                # null model, we simply use the first-order transition probabilities, i.e. P(b->c).             
                # Note that transition_matrices are transposed (!)
                v_1, w_1 = g1_node_mapping[p[-2]], g1_node_mapping[p[-1]]
                p_vw = T[w_1, v_1]

                # We use first-order transition probabilities to create an expected frequency 
                # of paths of length k. For a path (a,b,c) we use the count of (a,b) and 
                # "distribute" it to all possible paths (a,b,*) according to the first-order 
                # transition probabilities of (b,*)
                expected_vw = paths.paths[k-1][p[:k]].sum() * p_vw

                self.add_edge(v, w, weight = _np.array([0, expected_vw]))

        # Compute degrees of freedom of models
        if k == 0:
            # for a zero-order model, we just fit node probabilities
            # (excluding the special 'start' node)
            # Since probabilities must sum to one, the effective degree
            # of freedom is one less than the number of nodes
            # This holds for both the paths and the ngrams model
            self.dof_paths = self.vcount() - 2
            self.dof_ngrams = self.vcount() - 2
        else:
            # for a first-order model, self is the first-order network
            if k == 1:
                g1 = self
                A = g1.adjacency_matrix(include_subpaths=True, weighted=False,
                                        transposed=True)

            # Degrees of freedom in a higher-order ngram model
            s = g1.vcount()

            # The degrees of freedom of the higher-order model, under the ngram
            # assumption
            self.dof_ngrams = (s ** k) * (s - 1)

            # For k>0, the degrees of freedom of a path-based model depend on
            # the number of possible paths of length k in the first-order network.
            # Since probabilities in each row must sum to one, the degrees
            # of freedom must be reduced by one for each k-order node
            # that has at least one possible transition.

            # (A**k).sum() counts the number of different paths of exactly length k
            # based on the first-order network, which corresponds to the number of
            # possible transitions in the transition matrix of a k-th order model.
            paths_k = (A ** k).sum()

            # For the degrees of freedom, we must additionally consider that
            # rows in the transition matrix must sum to one, i.e. we have to
            # subtract one degree of freedom for every non-zero row in the (null-model)
            # transition matrix. In other words, we subtract one for every path of
            # length k-1
            # that can possibly be followed by at least one edge to a path of length k

            # This can be calculated by counting the number of non-zero elements in the
            # vector containing the row sums of A**k
            non_zero = _np.count_nonzero((A ** k).sum(axis=0))

            # The degrees of freedom of the higher-order model, under the paths
            # assumption
            self.dof_paths = paths_k - non_zero


    @staticmethod
    def generate_possible_paths(network, k):
        """ Returns all paths of length k that can
        possibly exist in a given network """

        assert k > 0, 'This function only calculates possible paths of length k > 0'

        # start with edges, i.e. paths of length one
        possible_paths = list(network.edges.keys())

        # extend all of those paths by an edge k-1 times
        for _ in range(k - 1):
            E_new = list()
            for e1 in possible_paths:
                for e2 in network.edges:
                    if e1[-1] == e2[0]:
                        p = e1 + (e2[1],)
                        E_new.append(p)
            possible_paths = E_new
        
        return possible_paths


    def total_edge_weight(self):
        """ Returns the sum of all edge weights """
        if self.edges:
            return sum(e['weight'] for e in self.edges.values())
        return _np.array([0, 0])

    def model_size(self):
        """
        Returns the number of non-zero elements in the adjacency matrix
        of the higher-order model.
        """
        return self.adjacency_matrix().count_nonzero()

    def first_order_nodes(self):
        """
        Returns a set of nodes projected to a first-order network
        """
        nodes = set()
        for v in self.nodes:
            for w in self.higher_order_node_to_path(v):
                nodes.add(w)
        return nodes


    def higher_order_node_to_path(self, node):
        """Helper function that transforms a node in a higher-order network of order k
        into a corresponding path of length k-1. For a higher-order node 'a-b-c-d'
        this function will return ('a','b','c','d')

        Parameters
        ----------
        node: str
            The higher-order node to be transformed to a path.
            TODO: this function assumes that the separator is '-', but it is not sure
            TODO: that the user will use it.

        Returns
        -------
        tuple
        """
        return tuple(node.split(self.separator))

    def path_to_higher_order_nodes(self, path, k=None):
        """Helper function that transforms a path of first-order nodes into a 
        sequence of k-order nodes using the separator character of the 
        HigherOrderNetwork instance

        Parameters
        ----------
        path:
            the path tuple to turn into a sequence of higher-order nodes
        k: int
            the order of the representation to use (default: order of the
            HigherOrderNetwork instance)

        Returns
        -------
        list

        Examples
        --------

        Consider an example path (a,b,c,d) with a separator string '-'

        >>> from pathpy import Paths
        >>> path_tuple = ('a', 'b', 'c', 'd')
        >>> paths = Paths()
        >>> paths.add_path_tuple(path_tuple)
        >>> hon = HigherOrderNetwork(paths, separator='-')
        >>> hon.path_to_higher_order_nodes(path_tuple, k=1)
        ['a', 'b', 'c', 'd']
        >>> hon.path_to_higher_order_nodes(path_tuple, k=2)
        ['a-b', 'b-c', 'c-d']
        >>> hon.path_to_higher_order_nodes(path_tuple, k=3)
        ['a-b-c', 'b-c-d']
        """

        if k is None:
            k = self.order
        assert len(path) >= k, 'Error: Path length must be at least k'

        if k == 0 and len(path) == 1:
            return ['start', path[0]]

        return [self.separator.join(path[n:n + k]) for n in range(len(path) - k + 1)]

    def node_to_name_map(self):
        """Returns a dictionary that can be used to map node names to matrix/vector indices"""
        return {v: idx for idx, v in enumerate(self.nodes)}

    def degrees_of_freedom(self, assumption="paths"):
        """Calculates the degrees of freedom (i.e. number of parameters) of
        this k-order model. Depending on the modeling assumptions, this either
        corresponds to the number of paths of length k in the first-order network
        or to the number of all possible k-grams. The degrees of freedom of a model
        can be used to assess the model complexity when calculating, e.g., the
        Bayesian Information Criterion (BIC).

        Parameters
        ----------
        assumption: str
            if set to 'paths', for the degree of freedom calculation in the BIC, only
            paths in the first-order network topology will be considered. This is needed
            whenever we are interested in a modeling of paths in a given network topology.
            If set to 'ngrams' all possible n-grams will be considered, independent of
            whether they are valid paths in the first-order network or not. The 'ngrams'
            and the 'paths' assumption coincide if the first-order network is fully
            connected.

        Returns
        -------
        int
        """
        assert assumption in ['paths', 'ngrams'], 'Error: Invalid assumption'

        if assumption == 'paths':
            return self.dof_paths
        return self.dof_ngrams

    def higher_order_path_to_first_order(self, path):
        """Maps a path in the higher-order network to a path in the first-order network.
        As an example, the second-order path ('a-b', 'b-c', 'c-d') of length two is mapped
        to the first-order path ('a','b','c','d') of length four.
        In general, a path of length l in a network of order k is mapped to a path of
        length l+k-1 in the first-order network.

        Parameters
        ----------
        path: str
            The higher-order path that shall be mapped to the first-order network

        Returns
        -------
        tuple
        """
        p1 = self.higher_order_node_to_path(path[0])
        for x in path[1:]:
            p1 += (self.higher_order_node_to_path(x)[-1],)
        return p1

    def summary(self):
        """Returns a string summary of this higher-order
           network
        """
        summary_fmt = (
            'Higher-order network of order k = {order}\n'
            '\n'
            'Nodes:\t\t\t\t{vcount}\n'
            'Links:\t\t\t\t{ecount}\n'
            'Total weight (subpaths/longest paths):\t{sub_w}/{uni_w}\n'
        )
        summary = summary_fmt.format(
            order=self.order, vcount=self.vcount(), ecount=self.ecount(),
            sub_w=self.total_edge_weight()[0], uni_w=self.total_edge_weight()[1]
        )
        return summary


    def adjacency_matrix(self, include_subpaths=True, weighted=True, transposed=False):
        """Returns a sparse adjacency matrix of the higher-order network. By default,
        the entry corresponding to a directed link source -> target is stored in row s and
        column t and can be accessed via A[s,t].

        Parameters
        ----------
        include_subpaths: bool
            if set to True, the returned adjacency matrix will account for the occurrence
            of links of order k (i.e. paths of length k-1) as subpaths
        weighted: bool
            if set to False, the function returns a binary adjacency matrix.
            If set to True, adjacency matrix entries will contain the weight of an edge.
        transposed: bool
            whether to transpose the matrix or not.

        Returns
        -------
        numpy cooc matrix
        """
        row = []
        col = []
        data = []

        node_to_coord = self.node_to_name_map()

        if transposed:
            for s, t in self.edges:
                row.append(node_to_coord[t])
                col.append(node_to_coord[s])
        else:
            for s, t in self.edges:
                row.append(node_to_coord[s])
                col.append(node_to_coord[t])

        # create array with non-zero entries
        if not weighted:
            data = _np.ones(len(self.edges.keys()))
        else:
            if include_subpaths:
                data = _np.array([float(self.edges[e]['weight'].sum()) for e in self.edges])
            else:
                data = _np.array([float(self.edges[e]['weight'][1]) for e in self.edges])

        shape = (self.vcount(), self.vcount())
        return _sparse.coo_matrix((data, (row, col)), shape=shape).tocsr()


    def transition_matrix(self, include_subpaths=True):
        """Returns a (transposed) random walk transition matrix corresponding to the
        higher-order network.

        Parameters
        ----------
        include_subpaths: bool
            whether or not to include subpath statistics in the transition probability
            calculation (default True)

        Returns
        -------

        """
        row = []
        col = []
        data = []
        # calculate weighted out-degrees (with or without subpaths)
        if include_subpaths:
            D = {n: self.nodes[n]['outweight'].sum() for n in self.nodes}
        else:
            D = {n: self.nodes[n]['outweight'][1] for n in self.nodes}

        node_to_coord = self.node_to_name_map()

        for (s, t) in self.edges:
            # either s->t has been observed as a longest path, or we are interested in
            # subpaths as well

            # the following makes sure that we do not accidentally consider zero-weight
            # edges (automatically added by default_dic)
            unique_weight = self.edges[(s, t)]['weight'][1]
            subpath_weight = self.edges[(s, t)]['weight'][0]
            is_valid = (unique_weight > 0 or (include_subpaths and subpath_weight > 0))
            if is_valid:
                row.append(node_to_coord[t])
                col.append(node_to_coord[s])
                if include_subpaths:
                    count = self.edges[(s, t)]['weight'].sum()
                else:
                    count = self.edges[(s, t)]['weight'][1]
                assert D[s] > 0, \
                    'Encountered zero out-degree for node "{s}" ' \
                    'while weight of link ({s}, {t}) is non-zero.'.format(s=s, t=t)
                prob = count / D[s]
                if prob < 0 or prob > 1:  # pragma: no cover
                    raise ValueError('Encountered transition probability {p} outside '
                                     '[0,1] range.'.format(p=prob))
                data.append(prob)

        data = _np.array(data)
        data = data.reshape(data.size, )

        shape = self.vcount(), self.vcount()
        return _sparse.coo_matrix((data, (row, col)), shape=shape).tocsr()


    def laplacian_matrix(self, include_subpaths=True):
        """
        Returns the transposed Laplacian matrix corresponding to the higher-order network.

        Parameters
        ----------
        include_subpaths: bool
            Whether or not subpath statistics shall be included in the calculation of
            matrix weights

        Returns
        -------

        """
        transition_matrix = self.transition_matrix(include_subpaths)
        identity_matrix = _sparse.identity(self.vcount())

        return identity_matrix - transition_matrix



#!/usr/bin/python -tt
# -*- coding: utf-8 -*-
# =============================================================================
# File      : helpers.py 
# Creation  : 24 Jan 2018
# Time-stamp: <Son 2018-05-06 11:06 juergen>
#
# Copyright (c) 2018 Jürgen Hackl <hackl@ibi.baug.ethz.ch>
#               http://www.ibi.ethz.ch
# $Id$ 
#
# Description : Some little helpers to make life easier
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>. 
# =============================================================================

from math import radians, cos, sin, asin, sqrt

from cnet import config,logger
log = logger(__name__)

def haversine(pos_1,pos_2):
    """
    Calculate the great circle distance between two points
    on the earth (specified in decimal degrees)
    """
    lat1, lon1  = pos_1
    lat2, lon2 = pos_2
    # convert decimal degrees to radians 
    lon1, lat1, lon2, lat2 = map(radians, [lon1, lat1, lon2, lat2])
    # haversine formula 
    dlon = lon2 - lon1 
    dlat = lat2 - lat1 
    a = sin(dlat/2)**2 + cos(lat1) * cos(lat2) * sin(dlon/2)**2
    c = 2 * asin(sqrt(a)) 
    # Radius of earth in kilometers is 6371
    km = 6371* c
    return km * 1000





def remove_duplicates(sequence):
    """ Removes duplicates from list, whilst preserving order

    Parameters
    ----------
    sequence: list

    Returns
    -------
    sequence : list
        A python list, where duplicate entries are removed

    Examples
    --------
    >>> removes_duplicates([2,3,5,5,1,1,7,8,8,2,2])
    [2,3,5,1,7,8,2]
    """
    seen = set()
    seen_add = seen.add
    return [x for x in sequence if not (x in seen or seen_add(x))]

def sort_large_file(filename):
    """ Sort large file with linux sort command (very fast) """
    # (head -n 1 paths.csv && tail -n +2 paths.csv | sort -t $',' -k 1) > paths_sorted.csv
    os.system("(head -n 1 "+filename+" && tail -n +2 "+filename+" | sort -t $',' -k 1) > "+filename)

def is_last_row(itr):
    old = next(itr)
    for new in itr:
        yield False, old
        old = new
    yield True, old



# =============================================================================
# eof
#
# Local Variables: 
# mode: python
# mode: linum
# End: 

 



#!/usr/bin/python -tt
# -*- coding: utf-8 -*-
# =============================================================================
# File      : shortest_path.py
# Creation  : 25 May 2018
# Time-stamp: <Son 2018-07-22 11:08 juergen>
#
# Copyright (c) 2018 Jürgen Hackl <hackl@ibi.baug.ethz.ch>
#               http://www.ibi.ethz.ch
# $Id$
#
# Description : Algorithms to find the shortest path between two nodes
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# =============================================================================

import heapq

from heapq import heappush, heappop
from itertools import count


from cnet import logger, Network, Path, Paths, Edge
from cnet.utils.exceptions import CnetError, CnetNotImplemented

from scipy.sparse import csgraph, find, isspmatrix
log = logger(__name__)


def shortest_path(network, source, sink, weight=None, mode='path', method='auto'):
    # check and format inputs
    adj, source, sink = _check_inputs(network, source, sink, weight=weight)

    # calculate shortest path
    try:
        cost, path = _shortest_path(adj, source, sink, method=method)
    except:
        cost = float('inf')
        path = []
        log.warn('No valid path was found!')
    # If network is a network object and type is path,
    # a Path object will be returned
    if isinstance(network, Network) and mode == 'path':
        p = Path(weight=cost)
        if len(path) > 0:
            for i in range(len(path)-1):
                u = network.nodes[path[i]].id
                v = network.nodes[path[i+1]].id
                e = network.edges[(u, v)]
                p.add_edge(e)
        return p
    elif isinstance(network, Network) and mode != 'path':
        path = [network.nodes[p].id for p in path]
        return cost, path
    else:
        return cost, path


def k_shortest_paths(network, source, sink, k, weight=None, mode='paths', method='auto'):
    # check inputs
    adj, source, sink = _check_inputs(network, source, sink, weight=weight)

    # calculate k shortest paths
    try:
        paths = _k_shortest_paths(adj, source, sink, k, method=method)
    except:
        paths = []
        log.warn('No valid paths were found!')

    # If network is a network object and type is path,
    # a Path object will be returned
    if isinstance(network, Network) and mode == 'paths':
        P = Paths()
        for path in paths:
            p = Path(weight=path[0])
            if len(path[1]) > 0:
                for i in range(len(path[1])-1):
                    u = network.nodes[int(path[1][i])].id
                    v = network.nodes[int(path[1][i+1])].id
                    e = network.edges[(u, v)]
                    p.add_edge(e)
                P.add_path(p)
        return P
    elif isinstance(network, Network) and mode != 'path':
        P = []
        for path in paths:
            cost = path[0]
            path = [network.nodes[int(p)].id for p in path[1]]
            P.append((cost, path))
        return P

    else:
        return paths


def _check_inputs(network, source, sink, weight=None):

    # check if a network object or a adjacency matrix is given
    if isinstance(network, Network):
        adj = network.adjacency_matrix(weight=weight)
    elif isspmatrix(network):
        adj = network
    else:
        log.error('Network is not defined properly! Only networks of type'
                  '<Network> or sparse scipy matrices can be handled.')
        raise CnetError

    # check source or sink nodes
    if isinstance(network, Network) and isinstance(source, str):
        source = network.nodes.index(source)
    if isinstance(network, Network) and isinstance(sink, str):
        sink = network.nodes.index(sink)
    if not isinstance(source, int) or not isinstance(sink, int):
        log.error('Nodes are not defined properly! Only the node ids or the '
                  'node indices are valid inputs!')
        raise CnetError

    return adj, source, sink


def _shortest_path(adj, source, sink, method='auto'):
    cost, predecessors = csgraph.shortest_path(adj, indices=source,
                                               method=method,
                                               return_predecessors=True)
    path = []
    i = sink
    while i != source:
        path.append(int(i))
        i = predecessors[i]
    path.append(source)
    return cost[sink], path[::-1]


def _k_shortest_paths(adj, source, sink, k, method='auto'):

    cost, sp = _shortest_path(adj, source, sink, method=method)

    top_k_path_set, paths, used, counter = [(cost, sp)], [sp], set(), 0
    for i in range(len(sp)-1):
        used.add((sp[i], sp[i+1]))

    # iteration
    for i in range(k-1):
        # get spur ring and rooting nodes
        for j in range(len(paths[counter])-1):
            root = paths[counter][j]
            root_cost = 0
            if j != 0:
                for r in range(j):
                    root_cost += adj[(paths[counter][r], paths[counter][r+1])]
            for n in find(adj[root])[1]:
                if (root, n) not in used:
                    spur_cost, spur_path = _shortest_path(
                        adj, n, sink, method=method)
                    p_cost = root_cost + adj[(root, n)] + spur_cost
                    p_path = paths[counter][:j+1]+spur_path
                    top_k_path_set.append((p_cost, p_path))
                    for p in range(len(p_path) - 1):
                        used.add((p_path[p], p_path[p+1]))
    sorted_paths = sorted(top_k_path_set)
    # return sorted_paths
    if k > len(sorted_paths):
        log.warn('Less then k={} paths were found!'.format(k))
        return sorted_paths
    else:
        return sorted_paths[:k]


# # My pure Python implementation
def dijkstra(network, source, sink, weight=None):
    queue, checked = [(0, source, [])], set()
    while queue:
        (cost, v, path) = heapq.heappop(queue)
        if v not in checked:
            path = path + [v]
            checked.add(v)
            if v == sink:
                return cost, path
            for e, o in network.nodes[v].heads:
                _e = network.edges[e]
                if o == 0:
                    _v = _e.v.id
                elif o == 1:
                    _v = _e.u.id
                heapq.heappush(queue, (cost+_e.weight(weight), _v, path))
    return float('inf'), path


def yen_k_sp(network, source, sink, k, weight=None):

    cost, sp = dijkstra(network, source, sink, weight=weight)

    n2e = network.nodes_to_edges_map()

    top_k_path_set, paths, used, counter = [(cost, sp)], [sp], set(), 0
    for i in range(len(sp)-1):
        used.add((sp[i], sp[i+1]))
    # iteration
    for i in range(k-1):
        # get spur ring and rooting nodes
        for j in range(len(paths[counter])-1):
            root = paths[counter][j]
            root_cost = 0
            if j != 0:
                for r in range(j):
                    root_cost += network.edges[n2e[(paths[counter]
                                                    [r], paths[counter][r+1])][0]].weight(weight)
            for e, o in network.nodes[root].heads:
                # get spur path and spur cost
                _e = network.edges[e]
                if o == 0:
                    n = _e.v.id
                elif o == 1:
                    n = _e.u.id
                if (root, n) not in used:
                    spur_cost, spur_path = dijkstra(
                        network, n, sink, weight=weight)
                    p_cost = root_cost + \
                        network.edges[n2e[(root, n)][0]].weight(
                            weight) + spur_cost
                    p_path = paths[counter][:j+1]+spur_path
                    top_k_path_set.append((p_cost, p_path))
                    for p in range(len(p_path) - 1):
                        used.add((p_path[p], p_path[p+1]))
    sorted_paths = sorted(top_k_path_set)
    # return sorted_paths
    if k > len(sorted_paths):
        return sorted_paths
    else:
        return sorted_paths[:k]


def ksp(network, source, sink, k=1, weight=None, mode='paths', method='auto'):

    # check if source is equal the sink node
    if source == sink:
        log.error('Source node {} is equal to the sink node '
                  '{}'.format(source, sink))
        raise CnetError()

    cost, sp = dijkstra(network, source, sink, weight=weight)

    # TODO: Check if there is a valid connection between two nodes.
    if sink not in sp:
        log.error('Node {} not reachable from {}'.format(sink, source))
        raise CnetError()

    # initialize variables
    costs = [cost]
    paths = [sp]
    c = count()
    B = []
    network_original = network.copy()

    # network.summary()
    for i in range(1, k):
        for j in range(len(paths[-1])-1):
            spur_node = paths[-1][j]
            root_path = paths[-1][:j + 1]

            edges_removed = []
            for c_path in paths:
                if len(c_path) > j and root_path == c_path[:j + 1]:
                    u = c_path[j]
                    v = c_path[j + 1]
                    if network.has_edge((u, v)):
                        _edge = network.edges[(u, v)]
                        network.remove_edge((u, v))
                        edges_removed.append(_edge)

            for n in range(len(root_path) - 1):
                node = root_path[n]
                heads = [e[0] for e in network.nodes[node].heads]
                for e in heads:
                    edges_removed.append(network.edges[e])
                network.remove_edges_from(heads)

            spur_path_cost, spur_path = dijkstra(
                network, spur_node, sink, weight=weight)

            if sink in spur_path:
                total_path = root_path[:-1] + spur_path
                root_path_cost = 0
                for i in range(len(root_path)-1):
                    root_path_cost += network_original.edges[(
                        root_path[i], root_path[i+1])].weight(weight)
                total_path_cost = root_path_cost + spur_path_cost
                heappush(B, (total_path_cost, next(c), total_path))

            for e in edges_removed:
                network.add_edge(e)

        if B:
            (l, _, p) = heappop(B)
            if p not in paths:
                costs.append(l)
                paths.append(p)
        else:
            break

    if isinstance(network, Network) and mode == 'paths':
        P = Paths()
        for i, path in enumerate(paths):
            p = Path(weight=costs[i])
            if len(path) > 0:
                for i in range(len(path)-1):
                    # u = network.nodes[int(path[1][i])].id
                    # v = network.nodes[int(path[1][i+1])].id
                    # e = network.edges[(u, v)]
                    e = network.edges[(path[i], path[i+1])]
                    p.add_edge(e)
                P.add_path(p)
        return P
    else:
        return costs, paths


# =============================================================================
# eof
#
# Local Variables:
# mode: python
# mode: linum
# mode: auto-fill
# fill-column: 80
# End:



# -*- coding: utf-8 -*-
#    pathpy is an OpenSource python package for the analysis of time series data
#    on networks using higher- and multi order graphical models.
#
#    Copyright (C) 2016-2017 Ingo Scholtes, ETH Zürich
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#    Contact the developer:

#    E-mail: ischoltes@ethz.ch
#    Web:    http://www.ingoscholtes.net

from collections import defaultdict
from pathpy.utils import Log, Severity
from pathpy.classes.network import Network


class DAG(Network):
    """
        Represents a directed acyclic graph (DAG) which
        can be used to generate pathway statistics.
    """

    def __init__(self, edges=None):
        """
        Constructs a directed acyclic graph from an edge list
        """

        super().__init__(directed=True)

        # Whether or not this graph is acyclic. None indicates that it is unknown
        self.is_acyclic = None

        # list of topologically sorted nodes
        self.sorting = []

        # Set of nodes with no incoming edges
        self.roots = set()

        # Set of nodes with no outgoing edges
        self.leafs = set()

        if edges is not None:
            self.add_edges(edges)

        # placeholder properties for topological sort
        self.parent = {}
        self.start_time = {}
        self.finish_time = {}
        self.edge_classes = {}
        self.top_sort_count = 0


    @classmethod
    def from_temporal_network(cls, tempnet, delta=1):
        """ Simple algorithm to create a directed acyclic graph from a 
            temporal network for a given maximum time difference delta

        Parameters
        ----------
        tempnet: pp.TemporalNetwork
            the temporal network instance to use in the DAG generation

        delta: int
            the maximum time difference t'- t between two time-stamped edges 
            (a,b,t) and (b,c,t') below which a causal path a -> b -> c will be 
            created in the DAG

        """
        dag = cls()

        # dictionary that maps time-unfolded nodes to actual nodes
        node_map = {}

        for (v, w, t) in tempnet.tedges:
            # create time-unfolded nodes v_t and w_{t+1}
            v_t = "{0}_{1}".format(v, t)
            node_map[v_t] = v

            # create one time-unfolded link for all delta in [1, delta]
            # this implies that for delta = 2 and an edge (a,b,1) two 
            # time-unfolded links (a_1, b_2) and (a_1, b_3) will be created
            for x in range(1, int(delta)+1):
                w_t = "{0}_{1}".format(w, t+x)                
                node_map[w_t] = w
                dag.add_edge(v_t, w_t)            

        return dag, node_map


    def add_edges(self, edges):
        """Add a list of edges

        Parameters
        ----------
        edges: list
            a list of edges [(s_1, t_1), (s_1, t_2), ...]

        """
        self_loops = 0
        redundant_edges = 0
        for e in edges:
            is_redundant = False
            has_self_loop = False
            if e[0] == e[1]:
                has_self_loop = True
                self_loops += 1
            if (e[0], e[1]) in self.edges:
                is_redundant = True
                redundant_edges += 1
            if not has_self_loop and not is_redundant:
                self.add_edge(e[0], e[1])
        if self_loops > 0:
            Log.add('Warning: omitted %d self-loops' % self_loops, Severity.WARNING)
        if redundant_edges > 0:
            Log.add('Warning: omitted %d redundant edges' % redundant_edges,
                    Severity.WARNING)

    def routes_from_node(self, v, node_mapping=None):
        """
        Constructs all paths from node v to any leaf nodes

        Parameters
        ----------
        v:
            node from which to start
        node_mapping: dict
            an optional mapping from node to a different set.

        Returns
        -------
        list
            a list of lists, where each list contains one path from the source
            node v until a leaf node is reached
        """
        # Collect temporary paths, indexed by the target node
        temp_paths = defaultdict(list)
        temp_paths[v] = [[v]]

        # set of unprocessed nodes
        queue = {v}

        while queue:
            # take one unprocessed node
            x = queue.pop()

            # successors of x expand all temporary
            # paths, currently ending in x
            if self.successors[x]:
                for w in self.successors[x]:
                    for p in temp_paths[x]:
                        temp_paths[w].append(p + [w])
                    queue.add(w)
                del temp_paths[x]

        # flatten list
        final_paths = []
        for possible_paths in temp_paths.values():
            for path in possible_paths:
                if node_mapping:
                    path = [node_mapping[k] for k in path]
                final_paths.append(path)

        return final_paths

    def routes_to_node(self, v, node_mapping=None):
        """
        Constructs all paths to node v from any root node

        Parameters
        ----------
        v:
            node from which to start
        node_mapping: dict
            an optional mapping from node to a different set.

        Returns
        -------
        list
            a list of lists, where each list contains one path from the source
            node v until a leaf node is reached
        """
        # Collect temporary paths, indexed by the target node
        temp_paths = defaultdict(list)
        temp_paths[v] = [[v]]

        # set of unprocessed nodes
        queue = {v}

        while queue:
            # take one unprocessed node
            x = queue.pop()

            # successors of x expand all temporary
            # paths, currently ending in x
            if self.predecessors[x]:
                for w in self.predecessors[x]:
                    for p in temp_paths[x]:
                        temp_paths[w].append(p + [w])
                    queue.add(w)
                del temp_paths[x]

        # flatten list
        final_paths = []
        for possible_paths in temp_paths.values():
            for path in possible_paths:
                path = list(reversed(path))
                if node_mapping:
                    path = [node_mapping[k] for k in path]
                final_paths.append(path)

        return final_paths

    def dfs_visit(self, v, parent=None):
        """Recursively visits nodes in the graph, classifying edges as (1) tree, (2)
        forward, (3) back or (4) cross edges.

        Parameters
        ----------
        v:
            node to be visited
        parent:
            the parent of this node (None for nodes) with no parents
        """
        self.parent[v] = parent
        self.top_sort_count += 1
        self.start_time[v] = self.top_sort_count
        if parent:
            self.edge_classes[(parent, v)] = 'tree'

        for w in self.successors[v]:
            if w not in self.parent:
                self.dfs_visit(w, v)
            elif w not in self.finish_time:
                self.edge_classes[(v, w)] = 'back'
                self.is_acyclic = False
            elif self.start_time[v] < self.start_time[w]:
                self.edge_classes[(v, w)] = 'forward'
            else:
                self.edge_classes[(v, w)] = 'cross'
        self.top_sort_count += 1
        self.finish_time[v] = self.top_sort_count
        self.sorting.append(v)

    def topsort(self):
        """
        Performs a topological sorting of the graph, classifying
        all edges as (1) tree, (2) forward, (3) back or (4) cross
        edges in the process.

        see Cormen 2001 for details
        """
        self.sorting = []
        self.parent = {}
        self.start_time = {}
        self.finish_time = {}
        self.edge_classes = {}
        self.top_sort_count = 0
        self.is_acyclic = True
        for v in self.nodes:
            if v not in self.parent:
                self.dfs_visit(v)
        self.sorting.reverse()

    def make_acyclic(self):
        """Removes all back-links from the graph to make it acyclic, then performs another
        topological sorting of the DAG
        """
        if self.is_acyclic is None:
            self.topsort()
        removed_links = 0
        if not self.is_acyclic:
            # Remove all back links
            for e in list(self.edge_classes):
                if self.edge_classes[e] == 'back':
                    self.remove_edge(*e)
                    removed_links += 1
            self.topsort()
            assert self.is_acyclic, "Error: make_acyclic did not generate acyclic graph!"
            Log.add('Removed ' + str(removed_links) +
                    ' back links to make graph acyclic', Severity.INFO)

    def summary(self):
        """
        Returns a string representation of this directed acyclic graph
        """

        summary = 'Directed Acyclic Graph'
        summary += '\n'
        summary += 'Nodes:\t\t' + str(len(self.nodes)) + '\n'
        summary += 'Roots:\t\t' + str(len(self.roots)) + '\n'
        summary += 'Leaves:\t\t' + str(len(self.leafs)) + '\n'
        summary += 'Links:\t\t' + str(len(self.edges)) + '\n'
        summary += 'Acyclic:\t' + str(self.is_acyclic) + '\n'
        return summary

    def isolate_nodes(self):
        """Return a set of isolated nodes"""
        return self.leafs & self.roots

    def __str__(self):
        """
        Returns the default string representation of this object
        """
        return self.summary()

    def add_edge(self, source, target, **kwargs):
        """
        Adds a directed edge to the graph
        """

        if source not in self.nodes:
            self.add_node(source, **kwargs)
            self.roots.add(source)
        if target not in self.nodes:
            self.add_node(target, **kwargs)
            self.leafs.add(target)

        self.leafs.discard(source)
        self.roots.discard(target)
        super().add_edge(source, target)
        self.is_acyclic = None

    def remove_edge(self, source, target):
        """remove an edge"""
        super().remove_edge(source, target)

        # if no more predecessors for the target then it becomes a root
        if not self.predecessors[target]:
            self.roots.add(target)

        # if no more successors for the source node then it becomes a leaf
        if not self.successors[source]:
            self.leafs.add(target)


    def write_file(self, filename, sep=','):
        """Writes a dag as an adjaceny list to file

        Parameters
        ----------
        filename
        sep

        Returns
        -------
        dag
        """
        with open(filename, 'w') as file:
            for edge in self.edges:
                file.write(sep.join(edge)+'\n')

    @classmethod
    def read_file(cls, filename, sep=',', maxlines=None, mapping=None, header=False):
        """
        Reads a directed acyclic graph from a file
        containing an edge list of the form

        source,target

        where ',' can be an arbitrary separator character
        """
        with open(filename, 'r') as f:
            edges = []

            if mapping is not None:
                Log.add('Filtering mapped edges')

            Log.add('Reading edge list ...')

            if header:  # Read header
                f.readline()
            for i, line in enumerate(f):
                if maxlines and i > maxlines:
                    break
                fields = line.rstrip().split(sep)
                try:
                    if mapping is None or (fields[0] in mapping and fields[1] in mapping):
                        edges.append((fields[0], fields[1]))

                except (IndexError, ValueError):  # pragma: no cover
                    msg = 'Ignoring malformed data in ' \
                          'line {}: "{}"'.format((i+header), line.strip())
                    Log.add(msg, Severity.WARNING)

        return cls(edges=edges)



#!/usr/bin/python -tt
# -*- coding: utf-8 -*-
# =============================================================================
# File      : setup.py
# Creation  : 21 May 2018
# Time-stamp: <Die 2018-05-29 07:49 juergen>
#
# Copyright (c) 2018 Jürgen Hackl <hackl@ibi.baug.ethz.ch>
#               http://www.ibi.ethz.ch
# $Id$
#
# Description : Setup script for network2tikz
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# =============================================================================
import os
from setuptools import setup, find_packages

about = {}
base_dir = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(base_dir, 'network2tikz', '__about__.py'), 'rb') as f:
    exec(f.read(), about)

with open('README.md') as f:
    readme = f.read()

with open('LICENSE') as f:
    license = f.read()

setup(
    name='network2tikz',
    version=about['__version__'],
    packages=find_packages(),
    url='https://github.com/hackl/network2tikz',
    download_url = 'https://pypi.org/project/network2tikz',
    author=about['__author__'],
    author_email=about['__email__'],
    install_requires=['numpy'],
    description='A converter that takes a network (cnet, igraph, networkx, pathpy, ...) and creates a tikz-network for smooth integration into LaTeX.',
    long_description = readme,
    long_description_content_type='text/markdown',
    license = about['__license__'],
    classifiers=[
        about['__status__'],
        about['__license__'],
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Intended Audience :: Science/Research',
        'Topic :: Multimedia :: Graphics :: Graphics Conversion',
        'Topic :: Scientific/Engineering :: Visualization',
        ]
)

# =============================================================================
# eof
#
# Local Variables:
# mode: python
# mode: linum
# mode: auto-fill
# fill-column: 80
# End:



#!/usr/bin/python -tt
# -*- coding: utf-8 -*-
# =============================================================================
# File      : plot.py 
# Creation  : 08 May 2018
# Time-stamp: <Son 2018-05-20 17:48 juergen>
#
# Copyright (c) 2018 Jürgen Hackl <hackl@ibi.baug.ethz.ch>
#               http://www.ibi.ethz.ch
# $Id$
#
# Description : Module to plot networks as tikz-network
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>. 
# =============================================================================

import os
import subprocess
import errno
import webbrowser

from cnet import logger
from cnet.utils.exceptions import CnetError
from cnet.visualization.drawing import TikzNetworkDrawer

log = logger(__name__)

def plot(network, filename=None,type=None,**kwds):
    """Plots the network as a tikz-network.

    This function generates a plot of the network, thereby different output can
    be chosen, including a standalone tex file, the tikz code, a compiled pdf or
    a node and edge list which also can be used with the tikz-network library.

    The plot function supports different network types. Currently supported are:

    * 'cnet',
    * 'networkx',
    * 'igraph',
    * 'pathpy'
    * node/edge list

    The appearance of the plot can be modified by keyword arguments which will
    be explained in more detail below. The arguments follow the options
    available in the tikz-network library.

    Note
    ----
    All the options are described in more detail in the tikz-library [tn]_.

    Parameters
    ----------

    network : network object
        Network to be drawn. The network can be a 'cnet', 'networkx', 'igraph',
        'pathpy' object, or a tuple of a node list and edge list.

    filename : file, string or None, optional (default = None)
        File or filename to save. The file ending specifies the
        output. i.e. is the file ending with '.tex' a tex file will be
        created; if the file ends with '.pdf' a pdf is created; if the file
        ends with '.csv', two csv files are generated (filename_nodes.csv
        and filename_edges.csv). If the filename is a tuple of strings, the
        first entry will be used to name the node list and the second entry
        for the edge list; and if no ending and no type is defined a
        temporary pdf file is compiled and shown.

    type : str or None, optional (default = None)
        Type of the output file. If no ending is defined trough the filename,
        the type of the output file can be specified by the type
        option. Currently the following output types are supported:
        'tex', 'pdf', 'csv' and 'dat'.

    kwds : keyword arguments, optional (default= no attributes)
        Attributes used to modify the appearance of the plot.
        For details see below.


    Keyword arguments used for the plotting:

    **Nodes:**

    - ``node_size`` : size of the node. The default is 0.6 cm.

    - ``node_color`` : color of the nodes. The default is light blue. Colors can
      be specified either by common color names, or by 3-tuples of floats
      (ranging between 0 and 255 for the R, G and B components).

    - ``node_opacity`` : opacity of the nodes. The default is 1. The range of the
      number lies between 0 and 1. Where 0 represents a fully transparent fill
      and 1 a solid fill.

    - ``node_label`` : labels drawn next to the nodes.

    - ``node_label_position`` : Per default the position of the label is in the
      center of the node. Classical Tikz commands can be used to change the
      position of the label. Instead, using such command, the position can be
      determined via an angle, by entering a number between -360 and 360. The
      origin (0) is the y axis. A positive number change the position counter
      clockwise, while a negative number make changes clockwise.

    - ``node_label_distance`` : distance between the node and the label.

    - ``node_label_color`` : color of the label.

    - ``node_label_size`` : font size of the label.

    - ``node_shape`` : shape of the vertices. Possibilities are:
      'circle', 'rectangle',  'triangle', and any other Tikz shape

    - ``node_style`` : Any other Tikz style option or command can be entered via
      the option style. Most of these commands can be found in the "TikZ and
      PGF Manual". Contain the commands additional options (e.g. shading =
      ball), then the argument for the style has to be between { } brackets.

    - ``node_layer`` : the node can be assigned to a specific layer.

    - ``node_label_off`` : is Boolean option which suppress all labels.

    - ``node_label_as_id`` : is a Boolean option which assigns the node id as label.

    - ``node_math_mode`` : is a Boolean option which transforms the labels into
      mathematical expressions without using the $ $ environment.

    - ``node_pseudo`` : is a Boolean option which creates a pseudo node, where only
      the node name and the node coordinate will be provided.

    **Edges:**

    - ``edge_width`` : width of the edges. The default unit is point (pt).

    - ``edge_color`` : color of the edges. The default is gray. Colors can
      be specified either by common color names, or by 3-tuples of floats
      (ranging between 0 and 255 for the R, G and B components).

    - ``edge_opacity`` : opacity of the edges. The default is 1. The range of the
      number lies between 0 and 1. Where 0 represents a fully transparent fill
      and 1 a solid fill.

    - ``edge_curved`` : whether the edges should be curved. Positive numbers
      correspond to edges curved in a counter-clockwise direction, negative
      numbers correspond to edges curved in a clockwise direction. Zero
      represents straight edges.

    - ``edge_label`` : labels drawn next to the edges.

    - ``edge_label_position`` : Per default the label is positioned in between
      both nodes in the center of the line. Classical Tikz commands can be used to
      change the position of the label.

    - ``edge_label_distance`` : The label position between the nodes can be
      modified with the distance option. Per default the label is centered
      between both nodes. The position is expressed as the percentage of the
      length between the nodes, e.g. of distance = 0.7, the label is placed at
      70% of the edge length away of Vertex i.

    - ``edge_label_color`` : color of the label.

    - ``edge_label_size`` : font size of the label.

    - ``edge_style`` : Any other Tikz style option or command can be entered via
      the option style. Most of these commands can be found in the "TikZ and
      PGF Manual". Contain the commands additional options (e.g. shading =
      ball), then the argument for the style has to be between { } brackets.

    - ``edge_arrow_size`` : arrow size of the edges.

    - ``edge_arrow_width`` : width of the arrowhead on the edge.

    - ``edge_loop_size`` :  modifies the length of the edge. The measure value has
      to be insert together with its units. Per default the loop size is 1 cm.

    - ``edge_loop_position`` : The position of the self-loop is defined via the
      rotation angle around the node. The origin (0) is the y axis. A positive
      number change the loop position counter clockwise, while a negative
      number make changes clockwise.

    - ``edge_loop_shape`` : The shape of the self-loop is defined by the enclosing
      angle. The shape can be changed by decreasing or increasing the argument
      value of the loop shape option.

    - ``edge_directed`` : is a Boolean option which transform edges to directed
      arrows. If the network is already defined as directed network this option
      is not needed, except to turn off the direction for one or more edges.

    - ``edge_math_mode`` : is a Boolean option which transforms the labels into
      mathematical expressions without using the $ $ environment.

    - ``edge_not_in_bg`` : Per default, the edge is drawn on the background layer
      of the tikz picture. I.e. objects which are created after the edges
      appear also on top of them. To turn this off, the option edge_not_in_bg
      has to be enabled.

    **General Options:**

    - ``units`` : string or tuple of strings, optional (default = ('cm','pt'))
      Per default, all size values are based on cm, and all line widths are
      defined in pt units. Whit this option the input units can be
      changed. Currently supported are: Pixel 'px', Points 'pt',
      Millimeters 'mm', and Centimeters 'cm'. If a single value is entered as
      unit all inputs have to be defined using this unit. If a tuple of units
      is given, the sizes are defined with the first entry the line widths with
      the second entry.

    - ``layout`` : dict
      A dictionary with the node positions on a 2-dimensional plane. The
      key value of the dict represents the node id while the value
      represents a tuple of coordinates (e.g. n = (x,y)). The initial
      layout can be placed anywhere on the 2-dimensional plane.

    - ``margins`` : None, int, float or dict, optional (default = None)
      The margins define the 'empty' space from the canvas border. If no
      margins are defined, the margin will be calculated based on the maximum
      node size, to avoid clipping of the nodes. If a single int or float is
      defined all margins using this distances. To define different the margin
      sizes for all size a dictionary with in the form of
      `{'top':2,'left':1,'bottom':2,'right':.5}` has to be used.

    - ``canvas`` : None, tuple of int or floats, optional (default = (6,6))
      Canvas or figure_size defines the size of the plot. The values entered as
      a tuple of numbers where the first number is width of the figure and the
      second number is the height of the figure. If the option ``units`` is not
      used the size is specified in cm. Per default the canvas is 6cm x 6cm. 

    - ``keep_aspect_ratio`` : bool, optional (default = True)
      Defines whether to keep the aspect ratio of the current layout. If
      ``False``, the layout will be rescaled to fit exactly into the
      available area in the canvas (i.e. removed margins). If ``True``, the
      original aspect ratio of the layout will be kept and it will be
      centered within the canvas.

    - ``standalone`` : bool, optional (default = True)
      If this option is true, a standalone latex file will be created. i.e. the
      figure can be compiled from this output file. If standalone is false,
      only the tikz environment is stored in the tex file, and can be imported
      in an existing tex file.

    - ``clean`` : bool, optional (default = True)
      Whether non-pdf files created that are created during compilation should
      be removed.

    - ``clean_tex`` : bool, optional (default = True)
      Also remove the generated tex file.

    - ``compiler`` : `str` or `None`, optional (default = None)
      The name of the LaTeX compiler to use. If it is None, cnet will choose a
      fitting one on its own. Starting with ``latexmk`` and then ``pdflatex``.

    - ``compiler_args`` : `list` or `None`, optional (default = None)
      Extra arguments that should be passed to the LaTeX compiler. If this is
      None it defaults to an empty list.

    - ``silent`` : bool, optional (default = True)
      Whether to hide compiler output or not.

    In the style dictionary multiple keywords can be used to address
    attributes. These keywords will be converted to an unique key word,
    used in the remaining code. This allows to keep the keywords used in
    'igrap'.

    ======= =================
    keys    other valid keys
    ======= =================
    node    vertex, v, n
    edge    link, l, e
    margins margin
    canvas  bbox, figure_size
    units   unit
    ======= =================

    Examples
    --------

    For illustration purpose a similar network as in the python-igrap tutorial
    is used. Instead of igraph, the cnet module is used for creating the
    network.

    Create an empty network object, and add some edges.

    >>> net = Network(name = 'my tikz test network',directed=True)
    >>> net.add_edges_from([('ab','a','b'), ('ac','a','c'), ('cd','c','d'),
    >>>                     ('de','d','e'), ('ec','e','c'), ('cf','c','f'),
    >>>                     ('fa','f','a'), ('fg','f','g'),('gg','g','g'),
    >>>                     ('gd','g','d')])

    Adding node and edge properties.

    >>> net.nodes['name'] = ['Alice', 'Bob', 'Claire', 'Dennis', 'Esther',
    >>>                      'Frank', 'George']
    >>> net.nodes['age'] = [25, 31, 18, 47, 22, 23, 50]
    >>> net.nodes['gender'] = ['f', 'm', 'f', 'm', 'f', 'm', 'm']
    >>> net.edges['is_formal'] = [False, False, True, True, True, False, True,
    >>>                           False, False, False]

    Already now the network can be plotted.

    >>> cn.plot(net)

    .. figure:: plot_01.png
       :alt: Simple plot with no visual styles
       :align: center
       :scale: 50

    Per default, the node positions are assigned uniform random. In order to
    create a layout, the layout methods of the network packages can be used. Or
    the position of the nodes can be directly assigned, in form of a dictionary,
    where the key is the node id and the value is a tuple of the node position
    in x and y.

    >>> layout = {'a': (4.3191, -3.5352), 'b': (0.5292, -0.5292),
    >>>           'c': (8.6559, -3.8008), 'd': (12.4117, -7.5239),
    >>>           'e': (12.7, -1.7069), 'f': (6.0022, -9.0323),
    >>>           'g': (9.7608, -12.7)}
    >>> plot(net,layout=layout)

    This should open an external pdf viewer showing a visual representation of
    the network, something like the one on the following figure:

    .. figure:: plot_02.png
       :alt: Simple plot with layout
       :align: center
       :scale: 50

    We can simply re-using the previous layout object here, but we also
    specified that we need a bigger plot (8 x 8 cm) and a larger margin around
    the graph to fit the self loop and potential labels (1 cm).

    >>> plot(net, layout=layout, canvas=(8,8), margin=1)

    .. figure:: plot_03.png
       :alt: Simple plot with layout and margins
       :align: center
       :scale: 50

    Note, instead of the command ``margins`` the command ``margin`` can be
    used. Also instead of ``canvas``, ``figure_size`` or ``bbox`` can be
    used. For more information see table above.

    In to keep the properties of the visual representation of your network
    separate from the network itself. You can simply set up a Python dictionary
    containing the keyword arguments you would pass to :py:meth:`plot` and then
    use the double asterisk (**) operator to pass your specific styling
    attributes to :py:meth:`plot`:

    >>> color_dict = {'m': 'blue', 'f': 'red'}
    >>> visual_style = {}

    Node options

    >>> visual_style['node_size'] = .5
    >>> visual_style['node_color'] = [color_dict[g] for g in net.nodes('gender')]
    >>> visual_style['node_opacity'] = .7
    >>> visual_style['node_label'] = net.nodes['name']
    >>> visual_style['node_label_position'] = 'below'

    Edge options

    >>> visual_style['edge_width'] = [1 + 2 * int(f) for f in net.edges('is_formal')]
    >>> visual_style['edge_curved'] = 0.1

    General options

    >>> visual_style["canvas"] = (8,8)
    >>> visual_style["margin"] = 1
    >>> plot(net,**visual_style)

    .. figure:: plot_04.png
       :alt: Plot with modified style
       :align: center
       :scale: 50

    Beside showing the network, we can also generate the latex source file,
    which can be used and modified later on. This is done by adding the output
    file name with the ending '.tex'

    >>> plot(net,'network.tex',**visual_style)

    .. code-block:: latex

       \\documentclass{standalone}
       \\usepackage{tikz-network}
       \\begin{document}
       \\begin{tikzpicture}
       \\clip (0,0) rectangle (8.0,8.0);
       \\Vertex[x=2.868,y=5.518,size=0.5,color=red,opacity=0.7,label=Alice,position=below]{a}
       \\Vertex[x=1.000,y=7.000,size=0.5,color=blue,opacity=0.7,label=Bob,position=below]{b}
       \\Vertex[x=5.006,y=5.387,size=0.5,color=red,opacity=0.7,label=Claire,position=below]{c}
       \\Vertex[x=6.858,y=3.552,size=0.5,color=blue,opacity=0.7,label=Dennis,position=below]{d}
       \\Vertex[x=7.000,y=6.419,size=0.5,color=red,opacity=0.7,label=Esther,position=below]{e}
       \\Vertex[x=3.698,y=2.808,size=0.5,color=blue,opacity=0.7,label=Frank,position=below]{f}
       \\Vertex[x=5.551,y=1.000,size=0.5,color=blue,opacity=0.7,label=George,position=below]{g}
       \\Edge[,lw=1.0,bend=-8.531,Direct](a)(b)
       \\Edge[,lw=1.0,bend=-8.531,Direct](a)(c)
       \\Edge[,lw=3.0,bend=-8.531,Direct](c)(d)
       \\Edge[,lw=3.0,bend=-8.531,Direct](d)(e)
       \\Edge[,lw=3.0,bend=-8.531,Direct](e)(c)
       \\Edge[,lw=1.0,bend=-8.531,Direct](c)(f)
       \\Edge[,lw=3.0,bend=-8.531,Direct](f)(a)
       \\Edge[,lw=1.0,bend=-8.531,Direct](f)(g)
       \\Edge[,lw=1.0,bend=-8.531,Direct](g)(g)
       \\Edge[,lw=1.0,bend=-8.531,Direct](g)(d)
       \\end{tikzpicture}
       \\end{document}

    Instead of the tex file, a node and edge list can be generates, which can
    also be used with the tikz-network library.

    >>> plot(net,'network.csv',**visual_style)

    The node list `network_nodes.csv`.

    .. code-block:: text

       id,x,y,size,color,opacity,label,position
       a,2.868,5.518,0.5,red,0.7,Alice,below
       b,1.000,7.000,0.5,blue,0.7,Bob,below
       c,5.006,5.387,0.5,red,0.7,Claire,below
       d,6.858,3.552,0.5,blue,0.7,Dennis,below
       e,7.000,6.419,0.5,red,0.7,Esther,below
       f,3.698,2.808,0.5,blue,0.7,Frank,below
       g,5.551,1.000,0.5,blue,0.7,George,below

    The edge list `network_edges.csv`.

    .. code-block:: text

       u,v,lw,bend,Direct
       a,b,1.0,-8.531,true
       a,c,1.0,-8.531,true
       c,d,3.0,-8.531,true
       d,e,3.0,-8.531,true
       e,c,3.0,-8.531,true
       c,f,1.0,-8.531,true
       f,a,3.0,-8.531,true
       f,g,1.0,-8.531,true
       g,g,1.0,-8.531,true
       g,d,1.0,-8.531,true

    References
    ----------

    .. [tn] https://github.com/hackl/tikz-network


    """
    result = Plot(network,**kwds)

    # get properties for the latex compiler
    _clean = kwds.get('clean',True)
    _clean_tex = kwds.get('clean_tex',True)
    _compiler = kwds.get('compiler',None)
    _compiler_args = kwds.get('compiler_arg',None)
    _silent = kwds.get('silent',True)

    if filename is None:
        filename = 'default_network'
    if isinstance(filename,tuple) or isinstance(filename,list) or \
         filename.endswith('.csv') or type == 'csv' or \
         filename.endswith('.dat') or type == 'dat':
        log.debug('Create csv files')
        result.save_csv(filename)
    elif filename == 'default_network' and type is None:
        log.debug('Show the network')
        # get current directory
        current_dir = os.getcwd()
        # create temp file name
        temp_filename = os.path.join(current_dir,filename)
        # save a pdf file
        result.save_pdf(filename, clean=_clean, clean_tex=_clean_tex,
                        compiler=_compiler, compiler_args=_compiler_args,
                        silent=_silent)
        # open the file
        webbrowser.open(r'file:///'+temp_filename+'.pdf')
        #result.show(filename)
    elif filename.endswith('.tex') or type == 'tex':
        log.debug('Create tex file')
        standalone = kwds.get('standalone',True)
        result.save_tex(filename,standalone=standalone)
    elif filename.endswith('.pdf') or type == 'pdf':
        log.debug('Create pdf plot')
        result.save_pdf(filename, clean=_clean, clean_tex=_clean_tex,
                        compiler=_compiler, compiler_args=_compiler_args,
                        silent=_silent)
    else:
        log.warn('No valid output option was chosen!')

class Plot(object):
    """Default class to create plots.

    The :py:class:`Plot` class is used to generate a network drawer and save the
    virtual network as a 'pdf', 'tex' or 'csv' file. Also a temporary file can
    be created to show the network figure.

    Parameters
    ----------
    network : network object
        Network to be drawn. The network can be a 'cnet', 'networkx', 'igraph',
        'pathpy' object, or a tuple of a node list and edge list.

    kwds : keyword arguments, optional (default= no attributes)
        Attributes to add to the drawer as key=value pairs.
        See also :py:meth:`plot`

    See Also
    --------
    plot

    """
    def __init__(self,network,**kwds):
        """Initialize the Plot class.

        Parameters
        ----------
        network : network object
            Network to be drawn. The network can be a 'cnet', 'networkx',
            'igraph', 'pathpy' object, or a tuple of a node list and edge list.

        kwds : keyword arguments, optional (default= no attributes)
            Attributes to add to the drawer as key=value pairs.
            See also :py:meth:`plot`

        """
        self.drawer = TikzNetworkDrawer(network,**kwds)

    def save_tex(self,filename,standalone=True):
        """Save the network as a tex file.

        Parameters
        ----------
        filename : file or string
            File or filename to save. File ending such as '.tex' is not
            necessary and will be added automatically if missing.

        standalone : bool, optional (default = True)
           If this option is true, a standalone latex file will be
           created. i.e. the figure can be compiled from this output file. If
           standalone is false, only the tikz environment is stored in the tex
           file, and can be imported in an existing tex file. 

        """
        latex_header = '\\documentclass{standalone}\n' + \
                        '\\usepackage{tikz-network}\n' + \
                        '\\begin{document}\n'
        latex_begin_tikz = '\\begin{tikzpicture}\n'
        latex_end_tikz = '\\end{tikzpicture}\n'
        latex_footer = '\\end{document}'

        # margins = self.drawer.margins
        w = self.drawer.canvas.width
        h = self.drawer.canvas.height
        latex_canvas = '\\clip (0,0) rectangle ({},{});\n'.format(w,h)

        # latex_margins = '\\fill[orange!20] ({},{}) rectangle ({},{});\n'\
        #                 ''.format(margins['left'],margins['bottom'],
        #                           canvas[0] - margins['right'],
        #                           canvas[1] - margins['top'])

        with open(filename,'w') as f:
            if standalone:
                f.write(latex_header)
            f.write(latex_begin_tikz)
            f.write(latex_canvas)
            #f.write(latex_margins)

            for node in self.drawer.node_drawer:
                f.write(node.draw())
            for edge in self.drawer.edge_drawer:
                f.write(edge.draw())

            f.write(latex_end_tikz)
            if standalone:
                f.write(latex_footer)

    def save_csv(self,filename):
        """Save the network as multiple csv files.

        Parameters
        ----------
        filename : file, string or tuple of files, strings.
            File or filename to save. If the filename is a string the output
            file names are a combination from the initial filename and '_nodes'
            or '_edges'. If the filename is a tuple of strings, the first string
            defines the output name for the node list and the second string
            defines the output name for the edge list.

        """
        # if file name is a string get base name
        if isinstance(filename,str):
            basename = os.path.splitext(os.path.basename(filename))[0]
            basename_n = basename + '_nodes'
            basename_e = basename + '_edges'
        # if the file name is a tuple, use the first part for the node list and
        # the second part for the edge list.
        elif isinstance(filename,tuple) or isinstance(filename,list):
            basename_n = os.path.splitext(os.path.basename(filename[0]))[0]
            basename_e = os.path.splitext(os.path.basename(filename[1]))[0]
        else:
            log.error('File name is not correct specified!')
            raise CnetError

        # write node list
        with open(basename_n+'.csv','w') as f:
            f.write(self.drawer.node_drawer[0].head())
            for node in self.drawer.node_drawer:
                f.write(node.draw(mode='csv'))

        # write edge list
        with open(basename_e+'.csv','w') as f:
            f.write(self.drawer.edge_drawer[0].head())
            for edge in self.drawer.edge_drawer:
                f.write(edge.draw(mode='csv'))

    def save_pdf(self,filename,clean=True, clean_tex=True,
                 compiler=None, compiler_args=None, silent=True):
        """Save the network as a tex file and compile the pdf.

        Note
        ----
        This code was adapted from the module `PyLaTeX`.

        Parameters
        ----------
        filename : file or string
            File or filename to save. File ending such as '.pdf' is not
            necessary and will be added automatically if missing.

        clean: bool, optional (default = True)
            Whether non-pdf files created that are created during compilation
            should be removed.

        clean_tex: bool, optional (default = True)
            Also remove the generated tex file.

        compiler: `str` or `None`, optional (default = None)
            The name of the LaTeX compiler to use. If it is None, cnet will
            choose a fitting one on its own. Starting with ``latexmk`` and then
            ``pdflatex``.

        compiler_args: `list` or `None`, optional (default = None)
            Extra arguments that should be passed to the LaTeX compiler. If
            this is None it defaults to an empty list.

        silent: bool, optional (default = True)
            Whether to hide compiler output or not.

        """
        if compiler_args is None:
            compiler_args = []

        # get directories and file name
        current_dir = os.getcwd()
        output_dir = os.path.dirname(filename)
        # check if output dir exists if not use the base dir
        if not os.path.exists(output_dir):
            output_dir = current_dir
        basename = os.path.splitext(os.path.basename(filename))[0]

        # change to output dir
        os.chdir(output_dir)

        # save the tex file
        self.save_tex(basename+'.tex',standalone=True)

        if compiler is not None:
            compilers = ((compiler, []),)
        else:
            latexmk_args = ['--pdf']

            compilers = (
                ('latexmk', latexmk_args),
                ('pdflatex', [])
            )

        main_arguments = ['--interaction=nonstopmode', basename + '.tex']

        os_error = None

        for compiler, arguments in compilers:
            command = [compiler] + arguments + compiler_args + main_arguments

            try:
                output = subprocess.check_output(command,
                                                 stderr=subprocess.STDOUT)
            except (OSError, IOError) as e:
                # Use FileNotFoundError when python 2 is dropped
                os_error = e

                if os_error.errno == errno.ENOENT:
                    # If compiler does not exist, try next in the list
                    continue
                raise
            except subprocess.CalledProcessError as e:
                # For all other errors print the output and raise the error
                print(e.output.decode())
                raise
            else:
                if not silent:
                    print(output.decode())

            if clean:
                try:
                    # Try latexmk cleaning first
                    subprocess.check_output(['latexmk', '-c', basename],
                                            stderr=subprocess.STDOUT)
                except (OSError, IOError, subprocess.CalledProcessError) as e:
                    # Otherwise just remove some file extensions.
                    extensions = ['aux', 'log', 'out', 'fls',
                                  'fdb_latexmk']

                    for ext in extensions:
                        try:
                            os.remove(basename + '.' + ext)
                        except (OSError, IOError) as e:
                            if e.errno != errno.ENOENT:
                                raise
            # remove the tex file
            if clean_tex:
                os.remove(basename + '.tex')
            # Compilation has finished, so no further compilers have to be tried
            break

        else:
            # Notify user that none of the compilers worked.
            log.error('No LaTex compiler was found! Either specify a LaTex '
                      'compiler or make sure you have latexmk or pdfLaTex'
                      ' installed.')
            raise CnetError

        # change back to current dir
        os.chdir(current_dir)


    def show(self,filename):
        """Show the compiled network.

        Create a tex file and compile the pdf out of it. After this is done the
        pdf will be automatically opened.

        Parameters
        ----------
        filename : file or string
            File or filename to save.

        See Also
        --------
        save_pdf

        """
        # get current directory
        current_dir = os.getcwd()
        # create temp file name
        temp_filename = os.path.join(current_dir,filename)
        # save a pdf file
        self.save_pdf(temp_filename)
        # open the file
        webbrowser.open(r'file:///'+temp_filename+'.pdf')


# =============================================================================
# eof
#
# Local Variables: 
# mode: python
# mode: linum
# mode: auto-fill
# fill-column: 80
# End:  



#!/usr/bin/python -tt
# -*- coding: utf-8 -*-
# =============================================================================
# File      : logger.py 
# Creation  : 30 Apr 2018
# Time-stamp: <Fre 2018-05-04 09:27 juergen>
#
# Copyright (c) 2018 Jürgen Hackl <hackl@ibi.baug.ethz.ch>
#               http://www.ibi.ethz.ch
# $Id$ 
#
# Description : Module to log output information
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>. 
# =============================================================================

import os
import logging
from cnet import config

# check if logging is enabled in the config file
if config.logging.enabled:
    # check if the path to the logging file exists
    if not os.path.exists(os.path.dirname(config.logging.logfile)):
        # if not create new path
        os.makedirs(os.path.dirname(config.logging.logfile))

    # initialize basic config for the written logger
    logging.basicConfig(level=logging._nameToLevel[config.logging.level],
                        format='%(asctime)s - %(name)12s:%(lineno)4d - %(levelname)8s - %(message)s',
                        datefmt='%Y-%m-%d %H:%M:%S',
                        filename=config.logging.logfile,
                        filemode='w')
else:
    # if logging is not enabled
    logging.basicConfig()

# check if logger should write to the terminal
if config.logging.verbose:

    # create stream handler
    console = logging.StreamHandler()

    # set level according to the config file
    console.setLevel(logging._nameToLevel[config.logging.level])

    # set a format which is simpler for console use
    formatter = logging.Formatter('[%(asctime)s: %(levelname)-5s] %(message)s',
                              datefmt='%m-%d %H:%M:%S')

    # tell the handler to use this format
    console.setFormatter(formatter)

    # add the handler to the root logger
    logging.getLogger('').addHandler(console)

def logger(name,level=None):
    """A function to generate logger for the modules."""

    # initialize new logger
    logger = logging.getLogger(name)

    # if no level is defined the config level will be used
    if level is None:
        logger.setLevel(logging._nameToLevel[config.logging.level])
    else:
        logger.setLevel(logging._nameToLevel[level])
    return logger




# =============================================================================
# eof
#
# Local Variables: 
# mode: python
# mode: linum
# mode: auto-fill
# fill-column: 80
# End:  



#!/usr/bin/python -tt
# -*- coding: utf-8 -*-
# =============================================================================
# File      : __init__.py 
# Creation  : 04 May 2018
# Time-stamp: <Fre 2018-05-04 09:26 juergen>
#
# Copyright (c) 2018 Jürgen Hackl <hackl@ibi.baug.ethz.ch>
#               http://www.ibi.ethz.ch
# $Id$ 
#
# Description : init file
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>. 
# =============================================================================





# =============================================================================
# eof
#
# Local Variables: 
# mode: python
# mode: linum
# mode: auto-fill
# fill-column: 80
# End:  



# -*- coding: utf-8 -*-

#    pathpy is an OpenSource python package for the analysis of time series data
#    on networks using higher- and multi order graphical models.
#
#    Copyright (C) 2016-2017 Ingo Scholtes, ETH Zürich
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#    Contact the developer:

#    E-mail: ischoltes@ethz.ch
#    Web:    http://www.ingoscholtes.net
from collections import defaultdict

import numpy as _np
import scipy.sparse.linalg as _sla

from pathpy import Network
from pathpy.utils import Log, Severity
from pathpy.utils import PathpyError

__all__ = ['connected_components']


def connected_components(network, lanczos_vecs=15, maxiter=1000):

    L = network.laplacian_matrix(weighted=True)
    n = network.vcount()-2
    vals, vecs = _sla.eigs(L, k=n, which="SM", ncv=lanczos_vecs, maxiter=maxiter, return_eigenvectors=True)

    components = defaultdict(set)
    c = 0

    # use eigenvectors of zero eigenvalues to map nodes to components
    for i in range(n):
        if _np.isclose(vals[i], 0, atol=1.e-12):
            print(vecs[:,i])
            min = _np.min(vecs[:,i])
            for i in _np.where(_np.isclose(vecs[:,i], min))[0]:
                components[c].add(i)
            c += 1
    return components



def reduce_to_gcc(network):
    """
    Returns a list of instances of the class Network that represent the 
    (strongly) connected components of a network. Connected components
    are calculated using Tarjan's algorithm.
    """

    # these are used as nonlocal variables (!)
    index = 0
    S = []
    indices = defaultdict(lambda: None)
    low_link = defaultdict(lambda: None)
    on_stack = defaultdict(lambda: False)
    components = {}

    # Tarjan's algorithm
    def strong_connect(v):
        nonlocal index
        nonlocal S
        nonlocal indices
        nonlocal low_link
        nonlocal on_stack
        nonlocal components

        indices[v] = index
        low_link[v] = index
        index += 1
        S.append(v)
        on_stack[v] = True

        for w in network.successors[v]:
            if indices[w] is None:
                strong_connect(w)
                low_link[v] = min(low_link[v], low_link[w])
            elif on_stack[w]:
                low_link[v] = min(low_link[v], indices[w])

        # create component of node v
        if low_link[v] == indices[v]:
            components[v] = set()
            while True:
                w = S.pop()
                on_stack[w] = False
                components[v].add(w)
                if v == w:
                    break

    # compute strongly connected components    
    for v in network.nodes:
        if indices[v] is None:
            strong_connect(v)
            # print('node {v}, size = {n}, component = {component}'.format(v=v, component=components[v], n = len(components[v]) ))

    max_size = 0
    for v in components:
        if len(components[v]) > max_size:            
            scc = components[v]
            max_size = len(components[v])

    # Reduce higher-order network to SCC
    for v in list(network.nodes):
        if v not in scc:
            # TODO: LV with the new network implementation the procedure removes edges but the in-degree
            # TODO: and the out-degree dictionaries are not updated and are out of sync.
            network.remove_node(v)



#!/usr/bin/python -tt
# -*- coding: utf-8 -*-
# =============================================================================
# File      : config.py 
# Creation  : 30 Apr 2018
# Time-stamp: <Fre 2018-05-04 09:27 juergen>
#
# Copyright (c) 2018 Jürgen Hackl <hackl@ibi.baug.ethz.ch>
#               http://www.ibi.ethz.ch
# $Id$ 
#
# Description : Module to read and parse configuration files
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>. 
# =============================================================================

import os
import sys
import collections
import configparser

class DotDict(collections.OrderedDict):
    """
    A string-valued dictionary that can be accessed with the "." notation
    """
    def __getattr__(self, key):
        try:
            if self[key] == 'True':
                return True
            elif self[key] == 'False':
                return False
            elif self.is_number(self[key]):
                if self.is_int(self[key]):
                    return int(self[key])
                else:
                    return float(self[key])
            else:
                return self[key]
        except KeyError:
            raise AttributeError(key)

    def is_number(self, s):
        """Check if input is a float number."""
        try:
            float(s)
            return True
        except TypeError:
            return False
        except ValueError:
            return False

    def is_int(self,x):
        """Check if input is an integer number."""
        try:
            a = float(x)
            b = int(a)
        except ValueError:
            return False
        else:
            return a == b

# initialize config dictionary
config = DotDict()

# add the default config file
d = os.path.dirname
base = os.path.join(d(d(__file__)), 'config.cfg')

config.paths = [base,'config.cfg']

def read(*paths, **validators):
    """
    Load the configuration, make each section available in a separate dict.

    The configuration location is where the script is executed:
       - cnet.cfg

    If this file is missing, the fallback is the source code:
       - cnet/config.cfg

    Please note: settings in the site configuration file are overridden
    by settings with the same key names in the config.cfg.
    """
    paths = config.paths + list(paths)
    parser = configparser.SafeConfigParser()
    found = parser.read(os.path.normpath(os.path.expanduser(p)) for p in paths)
    if not found:
        raise IOError('No configuration file found in %s' % str(paths))
    config.found = found
    config.clear()
    for section in parser.sections():
        config[section] = sec = DotDict(parser.items(section))
        for k, v in sec.items():
            sec[k] = validators.get(k, lambda x: x)(v)

config.read = read

config.read()

# =============================================================================
# eof
#
# Local Variables: 
# mode: python
# mode: linum
# mode: auto-fill
# fill-column: 80
# End:  



#!/usr/bin/python -tt
# -*- coding: utf-8 -*-
# =============================================================================
# File      : spatialnetwork.py
# Creation  : 01 May 2018
# Time-stamp: <Mit 2018-07-25 12:38 juergen>
#
# Copyright (c) 2018 Jürgen Hackl <hackl@ibi.baug.ethz.ch>
#               http://www.ibi.ethz.ch
# $Id$
#
# Description : Basic class for spatially embedded networks
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# =============================================================================

from cnet.classes.network import Node, Edge, Network
from cnet.classes.paths import Path

from cnet import logger
from cnet.utils.helpers import haversine
from cnet.utils.exceptions import CnetError, CnetNotImplemented
log = logger(__name__)


class SpatialNetwork(Network):
    """ Base class for spatially embedded networks.

    A Network stores nodes and edges with optional data, or attributes.

    Instances of this class capture a network that can be directed, undirected,
    unweighted or weighted. Self loops and multiple (parallel) edges are
    allowed.

    Nodes are defined as :py:class:`SpatialNode` and edges are defined as
    :py:class:`SpatialEdge`

    Parameters
    ----------
    directed : Boole, optional  (default = True)
        Specifies if a network contains directed edges, i.e u->v or
        undirected edges i.d. u<->v. Per default the network is assumed to
        be directed.

    name : string, optional (default='')
        An optional name for the network.

    attr : keyword arguments, optional (default= no attributes)
        Attributes to add to the network as key=value pairs.

    See Also
    --------
    Network, RoadNetwork

    """

    def __init__(self, directed=True, **attr):
        """Initialize a network with direction, name and attributes.

        Parameters
        ----------
        directed : Boole, optional (default = True)
            Specifies if a network contains directed edges, i.e u->v or
            undirected edges i.d. u<->v. Per default the network is assumed
            to be directed.

        name : string, optional (default='')
            An optional name for the graph.

        attr : keyword arguments, optional (default= no attributes)
            Attributes to add to graph as key=value pairs.

        """
        # initialize parent class of the network
        Network.__init__(self, directed=directed, **attr)

        # Classes of the Node and Edge objects
        # TODO: Probably there is a better solution to have different Node and
        # Edge classes for different Network sub classes
        self._node_class()
        self._edge_class()

    def _node_class(self):
        """Internal function to assign different Node classes."""
        self.NodeClass = SpatialNode

    def _edge_class(self):
        """Internal function to assign different Edge classes."""
        self.EdgeClass = SpatialEdge


class SpatialEdge(Edge):
    """Base class for an edge.

    Parameters
    ----------
    e : edge id
        The parameter e is the identifier (id) for the edge. Every edge should
        have a unique id. The id is converted to a string value and is used as a
        key value for all dict which saving edge objects.

    u : node id or py:class:`SpatialNode`
        This parameter defines the origin of the edge (if directed), i.e. u->v.
        A node id can be entered, in this case a new Node will be created
        (see py:class:`SpatialNode`), also an existing node can be used, here the
        attributes and properties of the node objects are used.

    v : node id or py:class:`SpatialNode`
        This parameter defines the destination of the edge (if directed)
        i.e. u->v. A node id can be entered, in this case a new Node will be
        created (see py:class:`SpatialNode`), also an existing node can be used,
        here the attributes and properties of the node objects are used.

    p1 : tuple of floats, optional (default = (0,0))
        Coordinate of the origin node, if origin node is not a
        py:class:`SpatialNode` object.

    p2 : tuple of floats, optional (default = (0,0))
        Coordinate of the destination node, if destination node is not a
        py:class:`SpatialNode` object.

    geometry : list of coordinate tuples, optional (default = None)
        Describing the geometry of the edge between the origin and destination
        nodes. Start and endpoint can be similar to the coordinates of the
        origin destination nodes.

    attr : keyword arguments, optional (default= no attributes)
        Attributes to add to edge as key=value pairs.

    Attributes
    ----------
    id : str
        Unique identifier for the edge. This property can only be called and not
        set or modified!

    u : Node
        Origin node of the edge. This property can only be called and not set or
        modified!

    v : Node
        Destination node of the edge. This property can only be called and not
        set or modified!

    coordinates : list of coordinates
        List of coordinates, describing the shape of the edge, beginning with
        the origin node coordinate and ending with the destination node
        coordinate. This property can only be called and not set or modified!

    Examples
    --------
    Create an empty edge with out coordinates (not recommended but possible).

    >>> ab = cn.SpatialEdge('ab','a','b')
    [WARNING] No suitable coordinate was found for node "a"! Coordinate was set to "(0,0)"!
    [WARNING] No suitable coordinate was found for node "b"! Coordinate was set to "(0,0)"!

    Create a proper edge

    >>> ab =  cn.SpatialEdge('ab','a','b', p1=(0,0), p2=(4,0))
    >>> ab.coordinates
    [(0,0),(4,0)]

    Create edge with predefined spatial nodes

    >>> c = cn.SpatialNode('c',x=2,y=4)
    >>> d = cn.SpatialNode('d',x=3,y=1)
    >>> cd = cn.SpatialEdge('cd',c,d)
    >>> cd.coordinates
    [(2,4),(3,1)]

    Create edge with geometry.

    >>> cd_geom = cn.SpatialEdge('cd',c,d,geometry=[(2,4),(4,3),(5,2)])
    >>> cd_geom.coordinates
    [(2,4),(4,3),(5,2),(3,1)]

    Calculating the euclidean length of the edge

    >>> cd.length()
    3.1622776601683795
    >>> cd_geom.length()
    5.8863495173726745

    See Also
    --------
    Edge, RoadEdge

    """

    def __init__(self, id, u, v, **attr):
        """Initialize the spatial edge object.

        Parameters
        ----------
        e : edge id
            The parameter e is the identifier (id) for the edge. Every edge
            should have a unique id. The id is converted to a string value and
            is used as a key value for all dict which saving edge objects.

        u : node id or py:class:`SpatialNode`
            This parameter defines the origin of the edge (if directed),
            i.e. u->v. A node id can be entered, in this case a new Node will be
            created (see py:class:`SpatialNode`), also an existing node can be
            used, here the attributes and properties of the node objects are
            used.

        v : node id or py:class:`SpatialNode`
            This parameter defines the destination of the edge (if directed)
            i.e. u->v. A node id can be entered, in this case a new Node will be
            created (see py:class:`SpatialNode`), also an existing node can be
            used, here the attributes and properties of the node objects are
            used.

        p1 : tuple of floats, optional (default = (0,0))
            Coordinate of the origin node, if origin node is not a
            py:class:`SpatialNode` object.

        p2 : tuple of floats, optional (default = (0,0))
            Coordinate of the destination node, if destination node is not a
            py:class:`SpatialNode` object.

        geometry : list of coordinate tuples, optional (default = None)
            Describing the geometry of the edge between the origin and
            destination nodes. Start and endpoint can be similar to the
            coordinates of the origin destination nodes.

        attr : keyword arguments, optional (default= no attributes)
            Attributes to add to edge as key=value pairs.

        Attributes
        ----------
        id : str
            Unique identifier for the edge. This property can only be called and
            not set or modified!

        u : Node
            Origin node of the edge. This property can only be called and not
            set or modified!

        v : Node
            Destination node of the edge. This property can only be called and
            not set or modified!

        coordinates : list of coordinates
            List of coordinates, describing the shape of the edge, beginning
            with the origin node coordinate and ending with the destination node
            coordinate. This property can only be called and not set or
            modified!

        """
        # Class of the Node object
        # TODO: Probably there is a better solution to have different Node
        # classes for different Edges sub classes
        self._node_class()

        # check the type of node object and if additional coordinates are defined
        if not isinstance(u, self.NodeClass) and 'p1' in attr:
            u = self.NodeClass(u, coordinate=attr['p1'])
        if not isinstance(v, self.NodeClass) and 'p2' in attr:
            v = self.NodeClass(v, coordinate=attr['p2'])

        # initialize the parent edge class
        Edge.__init__(self, id, u, v, **attr)

    def _node_class(self):
        """Internal function to assign different Node classes to the edge"""
        self.NodeClass = SpatialNode

    @property
    def coordinates(self):
        """Returns the associated coordinates of the edge.

        If an attribute geometry exist, the coordinates are extended with the
        coordinates stored in the the geometry attribute. Thereby, the origin
        coordinate is always at the first position and the destination
        coordinate is always at the last position.

        Returns
        -------
        coordinates : list of coordinate tuples
            Returns a list with all coordinates associated with the edge.

        """
        if 'geometry' in self.attributes:
            _coordinates = [self.u.coordinate]
            _coordinates += [c for c in self.attributes['geometry']]
            _coordinates += [self.v.coordinate]
            if _coordinates[0] == _coordinates[1]:
                _coordinates.pop(0)
            if _coordinates[-2] == _coordinates[-1]:
                _coordinates.pop(-1)
            return _coordinates
        else:
            return [self.u.coordinate, self.v.coordinate]

    def length(self, coordinates_type='euclidean'):
        """Returns the length of the edge.

        Parameters
        ----------
        coordinates_type : string, optional (default = 'euclidean')
            This parameter defines how the length of the edge is calculated. Per
            default the euclidean distance is calculated. With the option
            'topological' the topological length 1 is returned. With the option
            'geographic' the geographical distance measured along the surface of
            the earth (GPS) is calculated. If the option is None the edge
            attribute  "length" will be returned.

        Returns
        -------
        lenght : float
            Returns the length of the edge.

        Examples
        --------
        >>> ab = SpatialEdge('ab','a','b',p1=(0,0),p2=(5,0),length=13)
        >>> ab.length()
        5.0
        >>> ab.length(None)
        13
        >>> ab.lenght('topological')
        1
        >>> ab.lenght('geographic')

        The geographical distance is calculated based on GPS coordinates, where
        x corresponds to the latitude and y to the longitude coordinate. The
        length will be returned in meters.

        >>> gps = SpatialEdge('uv','u','v',p1=(47.409589, 8.502555),
        >>>                   p2=(47.410344, 8.503037),length=13)
        91.45243924583434

        """
        length = 0
        if coordinates_type is None:
            try:
                length = self.attributes['length']
            except:
                log.error('No attribute with the name "length" is defined!')
                raise CnetError
        elif coordinates_type is 'topological':
            length = 1
        elif coordinates_type is 'euclidean':
            coordinates = [complex(p[0], p[1]) for p in self.coordinates]
            for i in range(1, len(coordinates)):
                length += abs(coordinates[i] - coordinates[i-1])
        elif coordinates_type is 'geographic':
            # [complex(p[0],p[1]) for p in self.coordinates]
            coordinates = self.coordinates
            for i in range(1, len(coordinates)):
                length += haversine(coordinates[i-1], coordinates[i])
        else:
            log.error('The coordinates type "{}" is not defined! Only'
                      '"topological", "euclidean" and "geographic" are valid '
                      'values to enter.'.format(coordinates_type))
            raise CnetNotImplemented

        return length


class SpatialNode(Node):
    """Base class for a spatially embedded node.

    Parameters
    ----------
    u : node id
        The parameter u is the identifier (id) for the node. Every node should
        have a unique id. The id is converted to a string value and is used as a
        key value for all dict which saving node objects.

    x : float, optional (default = 0.0)
        The attribute x describes the x position of the node.

    y : float, optional (default = 0.0)
        The attribute y describes the y position of the node.

    coordinate : tuple of x and y, optional (default = (0.0,0.0))
        This attribute describes the coordinate as a tuple of x and y value.

    attr : keyword arguments, optional (default= no attributes)
        Attributes to add to node as key=value pairs.

    Attributes
    ----------
    id : str
        Unique identifier for the node. This property can only be called and not
        set or modified!

    x : float
        x position of the node. This property can be called, modified and set.

    y : float
        y position of the node. This property can be called, modified and set.

    coordinate : tuple of floats
        Tuple in form of (x,y), where x and y describe the position of the
        node. This property can be called, modified and set.

    Examples
    --------
    Create an empty node with out coordinate (not recommended but possible).

    >>> u = cn.SpatialNode('u')
    [WARNING] No suitable coordinate was found for node "u"! Coordinate was set to "(0,0)"!

    Create a proper node

    >>> v = cn.SpatialNode('u',x=1,y=2)
    >>> v.coordinate
    (1,2)
    >>> v.x
    1
    >>> v.y
    2

    Changing the coordinate

    >>> v.x = 3
    >>> v.coordinate
    (3,2)
    >>> v.coordinate = (4,4)
    >>> v.coordinate
    (4,4)

    See Also
    --------
    Node, RoadNode

    """

    def __init__(self, u, **attr):
        """Initialize the node object.

        Parameters
        ----------
        u : node id
            The parameter u is the identifier (id) for the node. Every node
            should have a unique id. The id is converted to a string value and
            is used as a key value for all dict which saving node objects.

        x : float, optional (default = 0.0)
            The attribute x describes the x position of the node.

        y : float, optional (default = 0.0)
            The attribute y describes the y position of the node. 

        coordinate : tuple of x and y, optional (default = (0.0,0.0))
            This attribute describes the coordinate as a tuple of x and y value.

        attr : keyword arguments, optional (default= no attributes)
            Attributes to add to node as key=value pairs.

        Attributes
        ----------
        id : str
            Unique identifier for the node. This property can only be called and
            not set or modified!

        x : float
            x position of the node. This property can be called, modified and
            set.

        y : float
            y position of the node. This property can be called, modified and
            set.

        coordinate : tuple of floats
            Tuple in form of (x,y), where x and y describe the position of the
            node. This property can be called, modified and set.

        """
        # initialize parent class
        Node.__init__(self, u, **attr)

        # check if attributes x and y are defined
        if 'x' in self.attributes and 'y' in self.attributes:
            self.coordinate = (self.attributes['x'], self.attributes['y'])
        elif 'x' in self.attributes and not 'y' in self.attributes:
            self.coordinate = (self.attributes['x'], 0)
        elif not 'x' in self.attributes and 'y' in self.attributes:
            self.coordinate = (0, self.attributes['y'])
        elif 'coordinate' in self.attributes:
            # TODO: add a check if coordinate is a tuple of floats
            None
        else:
            log.warn('No suitable coordinate was found for node "{}"!'
                     ' Coordinate was set to "(0,0)"!'.format(self.id))
            self.attributes['coordinate'] = (0.0, 0.0)

    @property
    def x(self):
        """Returns the x coordinate of the node."""
        return self.attributes['coordinate'][0]

    @x.setter
    def x(self, x):
        """Change the x coordinate of the node."""
        self.attributes['coordinate'] = (float(x), self.y)

    @property
    def y(self):
        """Returns the y coordinate of the node."""
        return self.attributes['coordinate'][1]

    @y.setter
    def y(self, y):
        """Change the y coordinate of the node."""
        self.attributes['coordinate'] = (self.x, float(y))

    @property
    def coordinate(self):
        """Returns the node coordinate as a tuple (x,y)."""
        return self.attributes['coordinate']

    @coordinate.setter
    def coordinate(self, coordinate):
        """Change the tuple (x,y) of the node coordinate."""
        self.attributes['coordinate'] = coordinate


class SpatialPath(Path, SpatialNetwork):
    """Base class for spatial embedded paths.

    A Path stores consecutive nodes and edges with optional data, or attributes.

    Instances of this class capture a path that can be directed, un-directed,
    un-weighted or weighted. The path class is a subclass of a
    :py:class:`SpatialNetwork`.

    Nodes are defined as :py:class:`SpatialNode` and edges are defined as
    :py:class:`SpatialEdge`

    Parameters
    ----------
    nodes : list of node ids or Nodes
        The parameter nodes must be a list with node ids or node
        objects. Every node within the list should have a unique id.
        The id is converted to a string value and is used as a key value for
        all dict which saving node objects.

    directed : Boole, optional  (default = True)
        Specifies if a network contains directed edges, i.e u->v or
        undirected edges i.d. u<->v. Per default the path is assumed to
        be directed.

    name : string, optional (default = '')
        An optional name for the path. If no name is assigned the network is
        called after the assigned nodes. e.g. if the path has nodes 'a', 'b' and
        'c', the network is named 'a-b-c'.

    separator : string, optional (default = '-')
        The separator used to separate the nodes in the node name.

    attr : keyword arguments, optional (default= no attributes)
        Attributes to add to the path as key=value pairs.

    Attributes
    ----------
    name : string
        Name of the path. If no name is assigned the network is called after the
        assigned nodes. e.g. if the path has nodes 'a', 'b' and 'c', the network
        is named 'a-b-c'. The maximum length of the name is defined by the
        constant `MAX_NAME_LENGTH`, which is per default 5. E.g. if the path has
        7 nodes (a,b,c,d,e,f,g) the name of the path is 'a-b-c-d-...-g'. Please,
        note the name of a path is NOT an unique identifier!

    Examples
    --------
    >>> p = cn.SpatialPath()
    >>> p.add_node(cn.SpatialNode('a', x=0, y=1))
    >>> p.add_node(cn.SpatialNode('b', x=2, y=3))

    See Also
    --------
    Path

    """

    def __init__(self, nodes=None, directed=True, separator='-', **attr):
        """Initialize a spatial path with direction, name and attributes.

        Parameters
        ----------
        nodes : list of node ids or Nodes
            The parameter nodes must be a list with node ids or node
            objects. Every node within the list should have a unique id.
            The id is converted to a string value and is used as a key value for
            all dict which saving node objects.

        directed : Boole, optional  (default = True)
            Specifies if a network contains directed edges, i.e u->v or
            undirected edges i.d. u<->v. Per default the path is assumed to
            be directed.

        name : string, optional (default = '')
            An optional name for the path. If no name is assigned the network is
            called after the assigned nodes. e.g. if the path has nodes 'a', 'b'
            and 'c', the network is named 'a-b-c'.

        separator : string, optional (default = '-')
            The separator used to separate the nodes in the node name.

        attr : keyword arguments, optional (default= no attributes)
            Attributes to add to the path as key=value pairs.

        """
        # initialize the parent class of the path
        Path.__init__(self, nodes=None, directed=True, separator='-', **attr)
        # initialize the parent class of the network
        SpatialNetwork.__init__(self, directed=directed, **attr)
        # assign nodes to the path
        if isinstance(nodes, list):
            super().add_nodes_from(nodes)


# =============================================================================
# eof
#
# Local Variables:
# mode: python
# mode: linum
# mode: auto-fill
# fill-column: 80
# End:



# -*- coding: utf-8 -*-
#    pathpy is an OpenSource python package for the analysis of time series data
#    on networks using higher- and multi order graphical models.
#
#    Copyright (C) 2016-2017 Ingo Scholtes, ETH Zürich
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#    Contact the developer:

#    E-mail: ischoltes@ethz.ch
#    Web:    http://www.ingoscholtes.net

from collections import defaultdict
import numpy as np

from pathpy.utils import Log


np.seterr(all='warn')


class MarkovSequence:
    """ Instances of this class can be used to fit
        standard higher-order Markov models for
        sequences generated from concatenated paths """

    def __init__(self, sequence):
        """
        Generates a Markov model for a sequence, given
        as a single list of strings
        """

        # The sequence to be modeled
        self.sequence = sequence

        # The transition probabilities of higher-order Markov chains
        self.P = {}

        # the set of states of higher-order Markov chains
        self.states = {1: set(sequence)}

    def fit_markov_model(self, k=1):
        """ Generates a k-th order Markov model
            for the underlying sequence
        """

        # TODO: Add support for k=0

        assert self.sequence, "Error: Empty sequence"

        # MLE fit of transition probabilities
        self.P[k] = defaultdict(lambda: defaultdict(lambda: 0.0))

        Log.add('Fitting Markov model with order k = ' + str(k))

        # Generate initial memory prefix
        mem = (())
        for s in self.sequence[:k]:
            mem += (s,)

        # count state transitions
        for s in self.sequence[k:]:
            self.P[k][mem][s] += 1.0

            # shift memory by one element
            mem = mem[1:] + (s,)

        # normalize transitions
        for m in self.P[k]:
            S = float(sum(self.P[k][m].values()))
            for s in self.P[k][m]:
                self.P[k][m][s] /= S
        Log.add('finished.')

    def likelihood(self, k=1, log=True):
        """
        Returns the likelihood of the sequence
        assuming a k-th order Markov model
        """

        if k not in self.P:
            self.fit_markov_model(k)

        L = 0

        # Generate initial prefix
        mem = (())
        for s in self.sequence[:k]:
            mem += (s,)

        for s in self.sequence[k:]:
            L += np.log(self.P[k][mem][s])

            # shift memory by one element
            mem = mem[1:] + (s,)

        if log:
            return L

        return np.exp(L)

    def bic(self, k=1, m=1):
        """ Returns the Bayesian Information Criterion
            assuming a k-th order Markov model """

        if k not in self.P:
            self.fit_markov_model(k)

        if m not in self.P:
            self.fit_markov_model(m)

        L_k = self.likelihood(k, log=True)
        L_m = self.likelihood(m, log=True)

        s = len(self.states[1])
        n = len(self.sequence)-k

        # the transition matrix of a first-order model with s states has s**2 entries,
        # subject to the constraint that entries in each row must sum up to one (thus
        # effectively reducing the degrees of freedom by a factor of s, i.e. we have
        # s^2-s^1. Generalizing this to order k, we arrive at
        # s^k * (s-1) = s^(k+1)-s^k degrees of freedom
        bic = np.log(n) * (s ** k - s ** m) * (s - 1) - 2.0 * (L_k - L_m)

        return bic

    def aic(self, k=1, m=1):
        """ Returns the Akaike Information Criterion
            assuming a k-th order Markov model """

        if k not in self.P:
            self.fit_markov_model(k)

        if m not in self.P:
            self.fit_markov_model(m)

        L_k = self.likelihood(k, log=True)
        L_m = self.likelihood(m, log=True)

        s = len(self.states[1])

        aic = 2 * (s**k - s**m) * (s-1) - 2.0 * (L_k - L_m)

        return aic

    def estimate_order(self, maxOrder, method='BIC'):
        """ Estimates the optimal order of a Markov model
            based on Likelihood, BIC or AIC """

        if method not in {'BIC', 'AIC', 'Likelihood'}:  # pragma: no cover
            msg = "Expecting method 'AIC', 'BIC' or 'Likelihood', got '%s'" % method
            raise ValueError(msg)

        values = []
        orders = []

        # We need k < m for the BIC and AIC calculation, which
        # is why we only test up to maxOrder - 1
        for k in range(1, maxOrder):
            if k not in self.P:
                self.fit_markov_model(k)

            orders.append(k)

            if method == 'AIC':
                values.append(self.aic(k, maxOrder))
            elif method == 'BIC':
                values.append(self.bic(k, maxOrder))
            elif method == 'Likelihood':
                values.append(self.likelihood(k, log=True))

        if method == 'Likelihood':
            values.append(self.likelihood(maxOrder, log=True))
            orders.append(maxOrder)

            # return order at which likelihood is maximized
            return orders[np.argmax(values)]

        # return order at which BIC/AIC are minimized
        return orders[np.argmin(values)]
